@@ -9,10 +9,10 @@ sampleWeights <- function(x, size, replace=FALSE, as.df=TRUE, as.points=FALSE, c
 		if (size >= nrow(res)) {
 			i <- 1:nrow(res)
 		} else {
-			i <- sample(nrow(res), size, prob=res[,ncol(res)], replace=replace)		
+			i <- sample.int(nrow(res), size, prob=res[,ncol(res)], replace=replace)
 		}
 	} else {
-		i <- sample(nrow(res), size, prob=res[,ncol(res)], replace=replace)
+		i <- sample.int(nrow(res), size, prob=res[,ncol(res)], replace=replace)
 	}
 	res <- res[i,]
 	if (as.points) {
@@ -28,76 +28,138 @@ sampleWeights <- function(x, size, replace=FALSE, as.df=TRUE, as.points=FALSE, c
 }
 
 
-sampleStratified <- function(x, size, replace=FALSE, as.df=TRUE, as.points=FALSE, cells=TRUE, xy=FALSE, ext=NULL, warn=TRUE, exp=2, weights=NULL) {
-	
-	if ((!xy) && (!as.points)) cells <- TRUE
-	
-	f <- freq(x)	
-	exp <- max(1, exp)
-	ss <- exp * size * nrow(f)
-	lonlat <- is.lonlat(x, perhaps=TRUE, warn=FALSE)
-	
-	if (is.null(weights)) {
-		if ((!lonlat) && (ss > (0.8 * ncell(x)))) { 
-			sr <- cbind(1:ncell(x), values(x))
-			colnames(sr) <- c("cell", names(x))
-		} else {
-			sr <- spatSample(x, ss, "random", replace=replace, na.rm=TRUE, ext=ext, cells=TRUE, values=TRUE, warn=warn)
+sampleStratMemory <- function(x, size, replace, lonlat, ext=NULL, weights=NULL, warn=TRUE) {
+	if (!is.null(ext)) {
+		x <- crop(x, ext)
+		if (!is.null(weights)) {
+			weights <- crop(weights, ext)
 		}
-	} else {
+	}
+	
+	if (!is.null(weights)) {
 		if (!inherits(weights, "SpatRaster")) {
-			error("spatSample", "weights must be a SpatRaster")			
+			error("spatSample", "weights must be a SpatRaster")
 		}
 		if (!compareGeom(x, weights)) {
 			error("spatSample", "geometry of weights does not match the geometry of x")
-		}	
-		sr <- vector("list", length = nrow(f))
-		for (i in 1:nrow(f)) {
-			r <- x == f[i,2]
-			r <- mask(weights, r, maskvalue=TRUE, inverse=TRUE)
-			sr[[i]] <- sampleWeights(r, size, replace=replace, cells=TRUE, ext=ext)[,1]
 		}
-		sr <- unlist(sr)
-		sr <- cbind(cell=sr, extract(x, sr)) 
+		v <- cbind(cell=1:ncell(x), values(x), values(weights))
+		v <- v[!is.na(v[,2]), ]
+		uv <- sort(unique(v[,2]))
+		ys <- vector(mode="list", length=length(uv))
+		for (i in seq_len(length(uv))) {
+			vv <- v[v[,2] == uv[i], ]
+			if (replace) {
+				s <- sample.int(nrow(vv), size, prob=vv[,3], replace=TRUE)
+			} else {
+				s <- sample.int(nrow(vv), min(size, nrow(vv)), prob=vv[,3], replace=FALSE)
+			}
+			ys[[i]] <- vv[s,-3]
+		}
+	} else if (lonlat) {
+		v <- cbind(cell=1:ncell(x), values(x), abs(cos(pi * values(init(x, "y")) / 360)))
+		v <- v[!is.na(v[,2]), ]
+		uv <- sort(unique(v[,2]))
+		ys <- vector(mode="list", length=length(uv))
+		for (i in seq_len(length(uv))) {
+			vv <- v[v[,2] == uv[i], ]
+			if (replace) {
+				s <- sample.int(nrow(vv), size, prob=vv[,3], replace=TRUE)
+			} else {
+				s <- sample.int(nrow(vv), min(size, nrow(vv)), prob=vv[,3], replace=FALSE)
+			}
+			ys[[i]] <- vv[s,-3]
+		}
+	} else {
+		v <- cbind(cell=1:ncell(x), values(x))
+		v <- v[!is.na(v[,2]), ]
+		uv <- sort(unique(v[,2]))
+		ys <- vector(mode="list", length=length(uv))
+		for (i in seq_len(length(uv))) {
+			vv <- v[v[,2] == uv[i], ]
+			if (replace) {
+				s <- sample.int(nrow(vv), size, replace=TRUE)
+			} else {
+				s <- sample.int(nrow(vv), min(size, nrow(vv)), replace=FALSE)
+			}
+			ys[[i]] <- vv[s,]
+		}
 	}
-	ys <- list()
-	notfound <- NULL
+	ys <- do.call(rbind, ys)
+	if (warn) {
+		ta <- tapply(ys[,1], ys[,2], length)
+		ta <- names(ta)[ta < size]
+		if (length(ta) > 0) {
+			warn("spatSample", 'fewer samples than requested are available for group(s): ', paste(ta, collapse=', '))
+		}
+	}
+	ys
+}
 
-	for (i in seq_len(nrow(f))) {
-		y <- sr[sr[, 2] == f[i,2], ,drop=FALSE]
-		if (nrow(y) == 0) {
-			notfound <- c(notfound, i)
+
+
+sampleStratified <- function(x, size, replace=FALSE, as.df=TRUE, as.points=FALSE, cells=TRUE, xy=FALSE, ext=NULL, warn=TRUE, exp=5, weights=NULL) {
+
+	lonlat <- is.lonlat(x, perhaps=TRUE, warn=FALSE)
+	if (blocks(x, n=4)$n == 1) {
+		res <- sampleStratMemory(x, size, replace, lonlat, ext, weights, warn)
+	} else {
+
+		#f <- freq(x)
+		f <- unique(x)[,1]
+		exp <- max(1, exp)
+		ss <- exp * size * length(f)
+
+		if (is.null(weights)) {
+			if ((!lonlat) && (ss > (0.8 * ncell(x)))) {
+				if (!is.null(ext)) {
+					x <- crop(x, ext)
+				}
+				sr <- cbind(1:ncell(x), values(x))
+				colnames(sr) <- c("cell", names(x))
+			} else {
+				sr <- spatSample(x, ss, "random", replace=replace, na.rm=TRUE, ext=ext, cells=TRUE, values=TRUE, warn=FALSE)
+			}
 		} else {
+			if (!inherits(weights, "SpatRaster")) {
+				error("spatSample", "weights must be a SpatRaster")
+			}
+			if (!compareGeom(x, weights)) {
+				error("spatSample", "geometry of weights does not match the geometry of x")
+			}
+			sr <- vector("list", length=length(f))
+			for (i in 1:length(f)) {
+				r <- x == f[i]
+				r <- mask(weights, r, maskvalue=TRUE, inverse=TRUE)
+				sr[[i]] <- sampleWeights(r, size, replace=replace, cells=TRUE, ext=ext)[,1]
+			}
+			sr <- unlist(sr)
+			sr <- cbind(cell=sr, extract(x, sr))
+		}
+		uv <- unique(sr[,2])
+		ys <- vector(mode="list", length=length(uv))
+		for (i in seq_len(length(uv))) {
+			y <- sr[sr[, 2] == uv[i], ,drop=FALSE]
 			if (nrow(y) > size) {
-				y <- y[sample(nrow(y), size),  ,drop=FALSE]
-			} 
+				y <- y[sample.int(nrow(y), size),  ,drop=FALSE]
+			}
 			ys[[i]] <- y
 		}
-	}
-	res <- do.call(rbind, ys)
-	colnames(res) <- c('cell', names(x))
-	
-	ures <- unique(res[,2])
-	miss <- !(f[,"value"] %in% ures)
-	if (any(miss) && warn) {
-		miss <- which(miss)
-		if (length(miss)== 1) {
-			warn("sample", 'no samples for stratum: ', miss)
-		} else if (length(miss) > 1) {
-			warn("sample", 'no samples for strata: ', paste(miss, collapse=', '))
-		}
-	}
-	
-	ta <- tapply(res[,1], res[,2], length) 
-	tanm <- names(ta)[which(ta < size)]
-	if ((length(tanm) > 0) && warn) {
-		if (length(tanm)== 1) {
-			warn("sample", 'fewer samples than requested for stratum: ', tanm)
-		} else if (length(tanm) > 1) {
-			warn("sample", 'fewer samples than requested for strata: ', paste(tanm, collapse=', '))
+		res <- do.call(rbind, ys)
+		colnames(res) <- c('cell', names(x))
+
+		if (warn) {
+			ta <- table(res[,2])
+			ta <- names(ta[ta < size])
+			tb <- f[!(f %in% unique(res[,2]))]
+			tba <- c(tb, ta)
+			if ((length(tba) > 0)) {
+				warn("spatSample", "fewer samples than requested for group(s): ", paste(tba, collapse=", "))
+			}
 		}
 	}
 
+	if ((!xy) && (!as.points)) cells <- TRUE
 	if (xy) {
 		pts <- xyFromCell(x, res[,1])
 		res <- cbind(res[,1,drop=FALSE], pts, res[,2,drop=FALSE])
@@ -112,87 +174,131 @@ sampleStratified <- function(x, size, replace=FALSE, as.df=TRUE, as.points=FALSE
 	}
 	if (!cells) {
 		res <- res[,-1,drop=FALSE]
-	}	
+	}
 	res
 }
 
 
-
-
-
-
-
-
 .seed <- function() {
   sample.int(.Machine$integer.max, 1)
 }
 
 
-.sampleCells <- function(x, size, method, replace, na.rm=FALSE, ext=NULL) {
+.sampleCellsMemory <- function(x, size, replace, lonlat, ext=NULL) {
+	if (!is.null(ext)) {
+		x <- crop(x, ext)
+	}
+	if (nlyr(x) > 1) {
+		x <- subst(anyNA(x), 1, NA)
+	}
+	if (lonlat) {
+		v <- cbind(cell=1:ncell(x), abs(cos(pi * values(init(x, "y")) / 360)), values(x))
+		v <- v[!is.na(v[,3]),]
+		i <- sample.int(nrow(v), min(size, nrow(v)), prob=v[,2], replace=replace)
+	} else {
+		v <- cbind(cell=1:ncell(x), values(x))
+		v <- v[!is.na(v[,2]),]
+		i <- sample.int(nrow(v), min(size, nrow(v)), replace=replace)
+	}
+	v[i,1]
+}
+
+
+.sampleCells <- function(x, size, method, replace, na.rm=FALSE, ext=NULL, exp=5) {
 	r <- rast(x)
 	lonlat <- is.lonlat(r, perhaps=TRUE, warn=TRUE)
 	if (!is.null(ext)) {
 		r <- crop(rast(r), ext)
 	}
-	if ( ((!replace) || (method == "regular")) && (size >= ncell(r)) ) {
-		cells <- 1:ncell(r)
-	}
 
-	if (method == "random") {
-		nsize <- size
+	if ((!replace) && (size >= ncell(r))) {
+		cells <- 1:ncell(r)
+	} else if (method == "random") {
 		if (na.rm) {
-			if (replace) {
-				size <- size*5
-			} else {
-				size <- min(ncell(r)*2, size*5)
-			}
+			esize <- size * exp
+		} else {
+			esize <- size
 		}
-		if (lonlat) {
+		if (na.rm && (blocks(x, n=4)$n == 1)) {
+			cells <- .sampleCellsMemory(x, size, replace, lonlat, ext)
+		} else if (lonlat) {
 			m <- ifelse(replace, 1.5, 1.25)
-			n <- m * size
+			n <- m * esize
 			y <- yFromRow(r, 1:nrow(r))
-			w <- abs(cos(pi*y/180))
-			rows <- sample.int(nrow(r), n, replace=TRUE, prob=w)
+			rows <- sample.int(nrow(r), n, replace=TRUE, prob=abs(cos(pi*y/180)))
 			cols <- sample.int(ncol(r), n, replace=TRUE)
 			cells <- cellFromRowCol(r, rows, cols)
 			if (!replace) {
 				cells <- unique(cells)
 			}
 		} else {
-			cells <- sample(ncell(r), size, replace=replace)
+			if (!replace) esize <- min(ncell(r), esize)
+			cells <- sample.int(ncell(r), esize, replace=replace)
 		}
-	} else { # regular 
-		if (lonlat) {
-			ratio <- 0.5 * ncol(r)/nrow(r)
-			n <- sqrt(size)
-			nx <- max(1, (round(n*ratio)))
-			ny <- max(1, (round(n/ratio)))
-			xi <- ncol(r) / nx
-			yi <- nrow(r) / ny
-			rows <- unique(round(seq(.5*yi, nrow(r), yi)))
-
-			w <- cos(pi*yFromRow(r, rows)/180)
-			w <- w * length(w)/sum(w)
-			xi <- xi / w
-			xi <- pmax(1,pmin(xi, ncol(r)))
-			z <- list()
-			#off <- stats::runif(1) 
-			for (i in 1:length(rows)) {
-				z[[i]] <- cbind(rows[i], unique(round(seq(0.5*xi[i], ncol(r), xi[i]))))
-			}
-			z <- do.call(rbind, z)
-			cells <- cellFromRowCol(r, z[,1], z[,2])
-
+	} else { # regular
+		if (TRUE) {
+			xy <- spatSample(ext(r), size, method, lonlat, FALSE)
+			cells <- cellFromXY(r, xy)
 		} else {
-			f <- sqrt(size / ncell(r))
-			nr <- ceiling(nrow(r) * f)
-			nc <- ceiling(ncol(r) * f);
-			xstep <- ncol(r) / nc
-			ystep <- nrow(r) / nr
-			xsamp <- seq(0.5*xstep, ncol(r), xstep)
-			ysamp <- seq(0.5*ystep, nrow(r), ystep)
-			xy <- expand.grid(ysamp, xsamp)
-			cells <- cellFromRowCol(r, xy[,1], xy[,2]) 
+			if (lonlat) {
+				#ratio <- ncol(r)/nrow(r)
+				e <- ext(r)
+				r1 = e$xmax - e$xmin;
+				r2 = e$ymax - e$ymin;
+				halfy = e$ymin + r2/2;
+
+				# beware that -180 is the same as 180; and that latitude can only go from -90:90 therefore:
+				dx = distance(cbind(e$xmin, halfy), cbind(e$xmin + 1, halfy), TRUE, TRUE) * min(180.0, r1);
+				dy = distance(cbind(0, e$ymin), cbind(0, e$ymax), TRUE, TRUE);
+				ratio = dy/dx;
+				n <- sqrt(size)
+				#nx <- max(1, (round(n*ratio)))
+				#ny <- max(1, (round(n/ratio)))
+				nx <- min(max(1, round(n/ratio)), ncol(r))
+				ny <- min(max(1, round(n*ratio)), nrow(r))
+
+				xi <- ncol(r) / nx
+				yi <- nrow(r) / ny
+				rows <- unique(round(seq(.5*yi, nrow(r), yi)))
+				w <- cos(pi*yFromRow(r, rows)/180)
+				w <- w * length(w)/sum(w)
+				xi <- xi / w
+				xi <- pmax(1,pmin(xi, ncol(r)))
+				z <- list()
+
+				# needs refinement:
+				global <- diff(e[1:2]) > 355
+
+				if (global) {
+					xi <- round(ncol(r) / round(ncol(r) / xi))
+					for (i in 1:length(rows)) {
+						if (xi[i] == 1) {
+							cols <- 1:ncol(r)
+						} else {
+							cols <- seq(xi[i]/2, ncol(r)-1, xi[i])
+						}
+						z[[i]] <- cbind(rows[i], cols)
+					}
+				} else {
+	#				xi <- round(ncol(r) / (round((ncol(r) / xi))+1))
+					for (i in 1:length(rows)) {
+						cols <- seq(xi[i]/2, ncol(r), xi[i])
+						z[[i]] <- cbind(rows[i], cols)
+					}
+				}
+				z <- do.call(rbind, z)
+				cells <- cellFromRowCol(r, z[,1], z[,2])
+			} else {
+				f <- sqrt(size / ncell(r))
+				nr <- ceiling(nrow(r) * f)
+				nc <- ceiling(ncol(r) * f);
+				xstep <- ncol(r) / nc
+				ystep <- nrow(r) / nr
+				xsamp <- seq(0.5*xstep, ncol(r), xstep)
+				ysamp <- seq(0.5*ystep, nrow(r), ystep)
+				xy <- expand.grid(ysamp, xsamp)
+				cells <- cellFromRowCol(r, xy[,1], xy[,2])
+			}
 		}
 	}
 	if (!is.null(ext)) {
@@ -203,15 +309,17 @@ sampleStratified <- function(x, size, replace=FALSE, as.df=TRUE, as.points=FALSE
 		cells <- cells[v]
 	}
 	if (method == "random") {
-		if (length(cells) > nsize) {
-			cells <- cells[1:nsize]
+		if (length(cells) > size) {
+			cells <- cells[1:size]
 		}
 	}
-	return(cells)
+	cells
 }
 
 
 set_factors <- function(x, ff, cts, asdf) {
+	if (!asdf) return(x)
+	
 	if (any(ff)) {
 		x <- data.frame(x)
 		for (i in which(ff)) {
@@ -223,16 +331,56 @@ set_factors <- function(x, ff, cts, asdf) {
 				x[[i]] <- ct[m,2]
 			}
 		}
-	} else if (asdf) {
-		x <- data.frame(x)
 	}
-	x
+	data.frame(x)
 }
 
 
-setMethod("spatSample", signature(x="SpatRaster"), 
-	function(x, size, method="random", replace=FALSE, na.rm=FALSE, as.raster=FALSE, as.df=TRUE, as.points=FALSE, values=TRUE, cells=FALSE, xy=FALSE, ext=NULL, warn=TRUE, weights=NULL) {
 
+.sampleCellsExhaustive <- function(x, size, replace, ext=NULL, weights=NULL, warn=TRUE) {
+
+	if (!is.null(ext)) {
+		x <- crop(x, ext)
+	}
+	rx <- rast(x)
+	x <- cells(x)
+	if (length(x) < size) {
+		if (!replace) {
+			warn("spatSample", "fewer samples than requested are available")
+			return(x)
+		}
+		size <- length(x)
+	}
+	
+	if (!is.null(weights)) {
+		if (!inherits(weights, "SpatRaster")) {
+			error("spatSample", "weights must be a SpatRaster")
+		}
+		weights <- weights[[1]]
+		if (!is.null(ext)) {
+			weights <- crop(weights, ext)
+		}
+		if (!compareGeom(x, weights)) {
+			error("spatSample", "geometry of weights does not match the geometry of x")
+		}
+		weights <- weights[x]
+		s <- sample.int(x, size, prob=weights, replace=replace)
+
+	} else if (is.lonlat(rx)) {	
+		y <- xyFromCell(rx, x)[,2]
+		weights <- abs(cos(pi * y / 360))
+		s <- sample(x, size, prob=weights, replace=replace)
+	} else {
+		s <- sample(x, size, replace=replace)
+	}
+	s
+}
+
+
+
+setMethod("spatSample", signature(x="SpatRaster"),
+	function(x, size, method="random", replace=FALSE, na.rm=FALSE, as.raster=FALSE, as.df=TRUE, as.points=FALSE, values=TRUE, cells=FALSE, xy=FALSE, ext=NULL, warn=TRUE, weights=NULL, exp=5, exhaustive=FALSE) {
+		exp <- max(c(1, exp), na.rm=TRUE)
 		size <- round(size)
 		if (any(size < 1)) {
 			error("spatSample", "sample size must be a positive integer")
@@ -244,30 +392,30 @@ setMethod("spatSample", signature(x="SpatRaster"),
 		method <- match.arg(tolower(method), c("random", "regular", "stratified", "weights"))
 		if (method == "stratified") {
 			if (as.raster) {
-				error("as.raster is not valid for method='stratified'")
+				error("spatSample", "as.raster is not valid for method='stratified'")
 			}
 			if (nlyr(x) > 1) {
 				x <- x[[1]]
-				warn("only the first layer of x is used")			
+				warn("spatSample", "only the first layer of x is used")
 			}
 			if (!hasValues(x)) {
-				error("x has no values")			
+				error("spatSample", "x has no values")
 			}
-			return( sampleStratified(x, size, replace=replace, as.df=as.df, as.points=as.points, cells=cells, xy=xy, ext=ext, warn=warn, exp=5, weights=weights) )
+			return( sampleStratified(x, size, replace=replace, as.df=as.df, as.points=as.points, cells=cells, xy=xy, ext=ext, warn=warn, exp=exp, weights=weights) )
 		} else if (!is.null(weights)) {
 			error("spatSample", "argument weights is only used when method='stratified'")
 		}
-		
+
 		if (method == "weights") {
 			if (as.raster) {
-				error("as.raster is not valid for method='weights'")
+				error("spatSample", "as.raster is not valid for method='weights'")
 			}
 			if (nlyr(x) > 1) {
 				x <- x[[1]]
-				warn("only the first layer of x is used")			
+				warn("spatSample", "only the first layer of x is used")
 			}
 			if (!hasValues(x)) {
-				error("x has no values")			
+				error("spatSample", "x has no values")
 			}
 			out <- try(sampleWeights(x, size, replace=replace, as.df=as.df, as.points=as.points, cells=cells, xy=xy, ext=ext) )
 			if (inherits(out, "try-error")) {
@@ -276,33 +424,39 @@ setMethod("spatSample", signature(x="SpatRaster"),
 			return (out)
 		}
 
-
 		if (!as.raster) {
 			ff <- is.factor(x)
-			lv <- active_cats(x)
+			lv <- levels(x)
 		}
 
 		if (cells || xy || as.points) {
+
 			size <- size[1]
-			cnrs <- .sampleCells(x, size, method, replace, na.rm, ext)
+			if (exhaustive && (method=="random") && na.rm) {
+				cnrs <- .sampleCellsExhaustive(x, size, replace, ext, weights=NULL, warn=FALSE)
+			} else {
+				cnrs <- .sampleCells(x, size, method, replace, na.rm, ext)
+			}
+			
 			if (method == "random") {
-				if (length(cnrs) < size) {
+				if (length(cnrs) < size && warn) {
 					warn("spatSample", "fewer cells returned than requested")
 				} else if (length(cnrs) > size) {
 					cnrs <- cnrs[1:size]
 				}
 			}
 			out <- NULL
+			
 			if (cells) {
 				out <- matrix(cnrs, ncol=1)
 				colnames(out) <- "cell"
 			}
+			
 			if (xy) {
 				out <- cbind(out, xyFromCell(x, cnrs))
 			}
 			if (values && hasValues(x)) {
 				e <- extract(x, cnrs)
-				e <- set_factors(e, ff, lv, as.df)
 				if (is.null(out)) {
 					out <- e
 				} else {
@@ -311,7 +465,8 @@ setMethod("spatSample", signature(x="SpatRaster"),
 			}
 			if (as.points) {
 				if (xy) {
-					out <- vect(out, crs=crs(x))
+					out <- data.frame(out)
+					out <- vect(out, geom=c("x", "y"), crs=crs(x))
 				} else {
 					xy <- xyFromCell(x, cnrs)
 					# xy is a matrix, no geom argument
@@ -339,8 +494,7 @@ setMethod("spatSample", signature(x="SpatRaster"),
 				} else {
 					x@ptr <- x@ptr$sampleRegularRaster(size)
 				}
-				x <- messages(x, "spatSample")
-				return(x);
+				return (messages(x, "spatSample"))
 			} else {
 				opt <- spatOptions()
 				if (length(size) > 1) {
@@ -358,10 +512,14 @@ setMethod("spatSample", signature(x="SpatRaster"),
 			}
 		} else { # random
 			size <- size[1]
+
 			if (as.raster) {
 				x@ptr <- x@ptr$sampleRandomRaster(size, replace, .seed())
 				x <- messages(x, "spatSample")
 				return(x);
+			} else if (exhaustive && na.rm) {
+				cnrs <- .sampleCellsExhaustive(x, size, replace, ext, weights=NULL, warn=FALSE)
+				out <- x[cnrs]	
 			} else {
 				#v <- x@ptr$sampleRandomValues(size, replace, seed)
 				if (size > 0.75 * ncell(x)) {
@@ -371,11 +529,11 @@ setMethod("spatSample", signature(x="SpatRaster"),
 						if (nrow(out) < size) {
 							warn("spatSample", "more non-NA cells requested than available")
 						} else {
-							out <- out[sample(nrow(out), size), ,drop=FALSE]
+							out <- out[sample.int(nrow(out), size), ,drop=FALSE]
 						}
 					} else {
 						out <- values(x)
-						out <- out[sample(nrow(out), size, replace=replace), ,drop=FALSE]
+						out <- out[sample.int(nrow(out), size, replace=replace), ,drop=FALSE]
 					}
 					out <- set_factors(out, ff, lv, as.df)
 					return(out)
@@ -385,11 +543,12 @@ setMethod("spatSample", signature(x="SpatRaster"),
 					scells <- NULL
 					ssize <- size*2
 					for (i in 1:10) {
-						scells <- c(scells, .sampleCells(x, ssize, method, replace))
+						scells <- c(scells, .sampleCells(x, ssize, method, replace, na.rm))
 						if ((i>1) && (!replace)) {
 							scells <- unique(scells)
 						}
-						out <- stats::na.omit(x[scells])
+						out <- extractCells(x, scells, raw=!as.df)   
+						out <- stats::na.omit(out)
 						if (nrow(out) >= size) {
 							out <- out[1:size, ,drop=FALSE]
 							attr(out, "na.action") <- NULL
@@ -399,17 +558,16 @@ setMethod("spatSample", signature(x="SpatRaster"),
 					}
 				} else {
 					scells <- .sampleCells(x, size, method, replace)
-					out <- x[scells]
+					out <- extractCells(x, scells, raw=!as.df)   
 				}
 				if (NROW(out) < size) {
 					if (warn) warn("spatSample", "fewer values returned than requested")
-				} else if (method == "random") {
-					if (is.null(dim(out))) {
-						out = out[1:size]
-					} else {
-						out = out[1:size, ,drop=FALSE]
-					}
+				} else if (is.null(dim(out))) {
+					out = out[1:size]
+				} else {
+					out = out[1:size, ,drop=FALSE]
 				}
+				out <- set_factors(out, ff, lv, as.df)
 				return(out)
 			}
 		}
@@ -417,7 +575,7 @@ setMethod("spatSample", signature(x="SpatRaster"),
 )
 
 
-setMethod("spatSample", signature(x="SpatExtent"), 
+setMethod("spatSample", signature(x="SpatExtent"),
 	function(x, size, method="random", lonlat, as.points=FALSE) {
 		if (missing(lonlat)) {
 			error("spatSample", "provide a lonlat argument")
@@ -451,7 +609,7 @@ setMethod("spatSample", signature(x="SpatExtent"),
 	cell <- cellFromXY(r, xy)
     uc <- unique(stats::na.omit(cell))
 
-	chess <- trim(chess)
+	chess <- trimws(chess)
 	if (chess != "") {
 		chess <- match.arg(tolower(chess), c("white", "black"))
 		nc <- ncol(r)
@@ -459,7 +617,7 @@ setMethod("spatSample", signature(x="SpatExtent"),
 			if (chess=="white") {
 				tf <- 1:ceiling(ncell(r)/2) * 2 - 1
 			} else {
-				tf <- 1:ceiling((ncell(r)-1)/2) * 2 
+				tf <- 1:ceiling((ncell(r)-1)/2) * 2
 			}
 		} else {
 			nr <- nrow(r)
@@ -520,9 +678,10 @@ get_field_name <- function(x, nms, sender="") {
 }
 
 
-setMethod("spatSample", signature(x="SpatVector"), 
+setMethod("spatSample", signature(x="SpatVector"),
 	function(x, size, method="random", strata=NULL, chess="") {
 		method = match.arg(tolower(method), c("regular", "random"))
+		size <- round(size)
 		stopifnot(size > 0)
 		gtype <- geomtype(x)
 		if (gtype == "polygons") {
@@ -532,7 +691,7 @@ setMethod("spatSample", signature(x="SpatVector"),
 						stopifnot(strata %in% names(x))
 					} else  {
 						stopifnot((strata > 0) && (strata < ncol(x)))
-					} 
+					}
 					strata <- x[[strata, drop=TRUE]]
 				} else if (length(strata) != length(x)) {
 					stop("length of strata must be 1 or length(x)")
@@ -540,7 +699,7 @@ setMethod("spatSample", signature(x="SpatVector"),
 				s <- stats::na.omit(unique(strata))
 				n <- length(size)
 				if (n==1) {
-					n <- rep_len(n, length(s)) 
+					n <- rep_len(n, length(s))
 				} else if (length(s) != n) {
 					stop("length of strata must be 1 or length(na.omit(unique(strata)))")
 				}
@@ -550,17 +709,19 @@ setMethod("spatSample", signature(x="SpatVector"),
 				r <- do.call(rbind, r)
 				return(r)
 			}
+			out <- vect()
 			if (length(size) == 1) {
-				x@ptr = x@ptr$sample(size, method[1], .seed())
+				out@ptr <- x@ptr$sample(size, method[1], .seed())
 			} else {
-				x@ptr = x@ptr$sampleGeom(size, method[1], .seed())
+				out@ptr <- x@ptr$sampleGeom(size, method[1], .seed())	
 			}
-			return(messages(x))
+			messages(x, "spatSample")
+			return(messages(out, "spatSample"))
 		} else if (grepl(gtype, "points")) {
 			if (!is.null(strata)) {
 				if (inherits(strata, "SpatRaster")) {
 					xy <- crds(x)
-					i <- .grid_sample(xy, size[1], rast(strata), chess) 
+					i <- .grid_sample(xy, size[1], rast(strata), chess)
 					return(x[i,])
 				} else {
 					error("spatSample", "not yet implemented for these strata")
@@ -578,7 +739,7 @@ setMethod("spatSample", signature(x="SpatVector"),
 
 
 
-# setMethod("spatSample", signature(x="SpatExtent"), 
+# setMethod("spatSample", signature(x="SpatExtent"),
 	# function(x, size, method="regular", lonlat, ...) {
 		# if (missing(lonlat)) {
 			# stop("provide a lonlat argument")
@@ -617,7 +778,7 @@ setMethod("spatSample", signature(x="SpatVector"),
 				# xi <- xi / w
 				# xi <- pmin(xi, 180)
 				# z <- list()
-				# #off <- stats::runif(1) 
+				# #off <- stats::runif(1)
 				# for (i in 1:length(lat)) {
 					# z[[i]] <- cbind(seq(e[1]+0.5*xi[i], e[2], xi[i]), lat[i])
 				# }
