@@ -7,13 +7,203 @@
 #include "gdal_alg.h"
 #include "ogrsf_frmts.h"
 
-#include "spatFactor.h"
+//#include "spatFactor.h"
 #include "recycle.h"
 #include "gdalio.h"
 
-SpatRaster rasterizePoints(SpatVector p, SpatRaster r, std::vector<double> values, double background, SpatOptions &opt) {
-	r.setError("not implemented in C++ yet");
-	return(r);
+
+SpatRaster SpatRaster::rasterizePoints(SpatVector x, std::string fun, std::vector<double> values, double background, SpatOptions &opt) {
+
+	SpatRaster out = geometry(1, false, false, false);
+	if (!out.writeStart(opt, filenames())) {
+		return out;
+	}
+	if (fun != "count" && (values.size() != x.size())) {
+		out.setError("values do not match geometries");
+		return out;
+	}
+
+	size_t nc = ncol();
+	std::vector<std::vector<double>> pxy = x.coordinates();
+	std::vector<double> cells = cellFromXY(pxy[0], pxy[1]);
+	if (fun == "count") {
+		for (size_t i=0; i < out.bs.n; i++) {
+			double cmin = out.bs.row[i] * nc;
+			double cmax = (out.bs.row[i]+out.bs.nrows[i]) * nc - 1;
+			std::vector<double> v(out.bs.nrows[i] * out.ncol(), 0);
+			for (size_t j=0; j<cells.size(); j++) {
+				if (cells[j] >= cmin && cells[j] <= cmax) {
+					size_t k = cells[j] - cmin;
+					v[k]++;
+				}
+			}
+			if (background != 0) {
+				for (size_t j=0; j<v.size(); j++) {
+					if (v[j] == 0) {
+						v[j] = background;
+					}
+				}
+			}
+			if (!out.writeValues(v, out.bs.row[i], out.bs.nrows[i]))  return out;
+		}
+	} else if (fun == "sum") {
+		for (size_t i=0; i < out.bs.n; i++) {
+			double cmin = out.bs.row[i] * nc;
+			double cmax = (out.bs.row[i]+out.bs.nrows[i]) * nc - 1;
+			std::vector<double> v(out.bs.nrows[i] * out.ncol(), 0);
+			for (size_t j=0; j<cells.size(); j++) {
+				if (cells[j] >= cmin && cells[j] <= cmax) {
+					size_t k = cells[j] - cmin;
+					v[k] += values[j];
+				}
+			}
+			if (background != 0) {
+				for (size_t j=0; j<v.size(); j++) {
+					if (v[j] == 0) {
+						v[j] = background;
+					}
+				}
+			}
+			if (!out.writeValues(v, out.bs.row[i], out.bs.nrows[i]))  return out;
+		}
+	} else if (fun == "mean") {
+		for (size_t i=0; i < out.bs.n; i++) {
+			double cmin = out.bs.row[i] * nc;
+			double cmax = (out.bs.row[i]+out.bs.nrows[i]) * nc - 1;
+			std::vector<double> v(out.bs.nrows[i] * out.ncol(), 0);
+			std::vector<double> cnt = v;
+			for (size_t j=0; j<cells.size(); j++) {
+				if (cells[j] >= cmin && cells[j] <= cmax) {
+					size_t k = cells[j] - cmin;
+					v[k] += values[j];
+					cnt[k]++;
+				}
+			}
+			for (size_t j=0; j<cnt.size(); j++) {
+				if (cnt[j] > 0) {
+					v[j] /= cnt[j];
+				}
+			}
+			if (background != 0) {
+				for (size_t j=0; j<v.size(); j++) {
+					if (v[j] == 0) {
+						v[j] = background;
+					}
+				}
+			}
+			if (!out.writeValues(v, out.bs.row[i], out.bs.nrows[i]))  return out;
+		}
+	} else { // "last"
+		for (size_t i=0; i < out.bs.n; i++) {
+			double cmin = out.bs.row[i] * nc;
+			double cmax = (out.bs.row[i]+out.bs.nrows[i]) * nc - 1;
+			std::vector<double> v(out.bs.nrows[i] * out.ncol(), background);
+			for (size_t j=0; j<cells.size(); j++) {
+				if (cells[j] >= cmin && cells[j] <= cmax) {
+					size_t k = cells[j] - cmin;
+					v[k] = values[j];
+				}
+			}
+			if (!out.writeValues(v, out.bs.row[i], out.bs.nrows[i]))  return out;
+		}
+	}
+	out.writeStop();
+	return out;
+}
+
+
+SpatRaster SpatRaster::rasterizeGeom(SpatVector x, std::string unit, std::string fun, SpatOptions &opt) {
+
+	if (x.type() != "points") {
+
+		SpatRaster out = geometry(1, false, false, false);
+		SpatOptions ops(opt);
+
+		std::vector<std::string> ss {"m", "km"};
+		if (std::find(ss.begin(), ss.end(), unit) == ss.end()) {
+			out.setError("invalid unit (not 'm' or 'km')");
+			return out;
+		}
+		if ((x.type() == "lines")) {
+			ss = {"count", "length", "crosses"};
+			if (std::find(ss.begin(), ss.end(), fun) == ss.end()) {
+				out.setError("invalid value for 'fun' (not 'count', 'crosses', or 'length')");
+				return out;
+			}
+		} else {
+			ss = {"area", "count"};
+			if (std::find(ss.begin(), ss.end(), fun) == ss.end()) {
+				out.setError("invalid value for 'fun' (not 'area' or 'count')");
+				return out;
+			}
+		}
+
+		SpatRaster empty = out.geometry();
+		SpatExtent e = out.getExtent();
+		double rsy = out.yres() / 2;
+
+		double m = unit == "m" ? 1 : 1000;
+		if (!x.is_lonlat()) {
+			double tom = x.srs.to_meter();
+			tom = std::isnan(tom) ? 1 : tom;
+			m *= tom;
+		}
+		if (x.type() == "lines") {
+			out.setNames({"length"});
+		} else {
+			out.setNames({"area"});
+			m *= m;
+		}
+		opt.ncopies = std::max(opt.ncopies, (unsigned)4) * 8;
+		if (!out.writeStart(opt, filenames())) {
+			return out;
+		}
+		for (size_t i=0; i < out.bs.n; i++) {
+			e.ymax = yFromRow(out.bs.row[i]) + rsy;
+			e.ymin = yFromRow(out.bs.row[i] + out.bs.nrows[i] - 1) - rsy;
+			SpatRaster tmp = empty.crop(e, "near", false, ops);
+
+			SpatVector p = tmp.as_polygons(true, false, false, false, false, ops);
+			std::vector<double> v(out.bs.nrows[i] * out.ncol(), 0);
+
+			if (fun == "crosses") {
+				std::vector<int> r = p.relate(x, "crosses", true, true);
+				size_t nx = x.size();
+				for (size_t j=0; j< r.size(); j++) {
+					size_t k= j / nx;
+					v[k] += r[j];
+				}
+			} else {
+				std::vector<long> cell(p.size());
+				std::iota(cell.begin(), cell.end(), 0);
+				p.df.add_column(cell, "cell");
+				p = p.intersect(x, true);
+				std::vector<double> stat;
+				if (x.type() == "lines") {
+					stat = p.length();
+				} else {
+					stat = p.area("m", false, {});
+				}
+				if (fun == "count") {
+					for (size_t j=0; j<stat.size(); j++) {
+						size_t k = p.df.iv[0][j];
+						v[k]++;
+					}
+				} else {
+					for (size_t j=0; j<stat.size(); j++) {
+						size_t k = p.df.iv[0][j];
+						v[k] += (stat[j] / m);
+					}
+				}
+			}
+			if (!out.writeValues(v, out.bs.row[i], out.bs.nrows[i]))  return out;
+		}
+		out.writeStop();
+		return(out);
+
+	} else {
+		return rasterizePoints(x, "count", {}, 0.0, opt);
+	}
 }
 
 
@@ -27,12 +217,12 @@ SpatRaster SpatRaster::hardCopy(SpatOptions &opt) {
 		out.setError(getError());
 		return(out);
 	}
- 	if (!out.writeStart(opt)) {
+ 	if (!out.writeStart(opt, filenames())) {
 		readStop();
 		return out;
 	}
 	for (size_t i = 0; i < out.bs.n; i++) {
-		std::vector<double> v; 
+		std::vector<double> v;
 		readBlock(v, out.bs, i);
 		if (!out.writeBlock(v, i)) return out;
 	}
@@ -60,14 +250,16 @@ SpatRaster SpatRaster::hardCopy(SpatOptions &opt) {
 bool SpatRaster::getDSh(GDALDatasetH &rstDS, SpatRaster &out, std::string &filename, std::string &driver, double &naval, bool update, double background, SpatOptions &opt) {
 
 	filename = opt.get_filename();
+	SpatOptions ops(opt);
+	ops.ncopies += 4;
 	if (filename == "") {
-		if (canProcessInMemory(opt)) {
+		if (canProcessInMemory(ops)) {
 			driver = "MEM";
 		} else {
 			filename = tempFile(opt.get_tempdir(), opt.pid, ".tif");
 			opt.set_filenames({filename});
 			driver = "GTiff";
-		} 
+		}
 	} else {
 		driver = opt.get_filetype();
 		getGDALdriver(filename, driver);
@@ -76,7 +268,7 @@ bool SpatRaster::getDSh(GDALDatasetH &rstDS, SpatRaster &out, std::string &filen
 			return false;
 		}
 		std::string msg;
-		if (!can_write(filename, opt.get_overwrite(), msg)) {
+		if (!can_write({filename}, filenames(), opt.get_overwrite(), msg)) {
 			out.setError(msg);
 			return false;
 		}
@@ -113,6 +305,7 @@ bool SpatRaster::getDSh(GDALDatasetH &rstDS, SpatRaster &out, std::string &filen
 
 SpatRaster SpatRaster::rasterizeLyr(SpatVector x, double value, double background, bool touches, bool update, SpatOptions &opt) {
 
+// not working well in some cases. See #552
 	std::string gtype = x.type();
 	SpatRaster out;
 	out.setNames({"ID"});
@@ -157,7 +350,7 @@ SpatRaster SpatRaster::rasterizeLyr(SpatVector x, double value, double backgroun
 	char** papszOptions = NULL;
 	CPLErr err;
 	if (touches) {
-		papszOptions = CSLSetNameValue(papszOptions, "ALL_TOUCHED", "TRUE"); 
+		papszOptions = CSLSetNameValue(papszOptions, "ALL_TOUCHED", "TRUE");
 	}
 	err = GDALRasterizeLayers(rstDS, static_cast<int>(bands.size()), &(bands[0]),
 			1, &(ahLayers[0]), NULL, NULL, &(values[0]), papszOptions, NULL, NULL);
@@ -188,13 +381,13 @@ SpatRaster SpatRaster::rasterizeLyr(SpatVector x, double value, double backgroun
 
 	GDALClose(rstDS);
 	if (driver != "MEM") {
-		out = SpatRaster(filename, {-1}, {""}, {});
+		out = SpatRaster(filename, {-1}, {""}, {}, {});
 	}
 	return out;
 }
 
 
-SpatRaster SpatRaster::rasterize(SpatVector x, std::string field, std::vector<double> values, 
+SpatRaster SpatRaster::rasterize(SpatVector x, std::string field, std::vector<double> values,
 	double background, bool touches, bool add, bool weights, bool update, bool minmax, SpatOptions &opt) {
 
 	std::string gtype = x.type();
@@ -204,12 +397,14 @@ SpatRaster SpatRaster::rasterize(SpatVector x, std::string field, std::vector<do
 	if (weights && ispol) {
 		SpatOptions sopts(opt);
 		SpatRaster wout = geometry(1);
+		field = "";
 		unsigned agx = 1000 / ncol();
-		agx = std::max((unsigned)10, agx); 
+		agx = std::max((unsigned)10, agx);
 		unsigned agy = 1000 / nrow();
 		agy = std::max((unsigned)10, agy);
+		//unsigned agx = 100;
+		//unsigned agy = 100;
 		wout = wout.disaggregate({agx, agy}, sopts);
-		field = "";
 		double f = agx * agy;
 		wout = wout.rasterize(x, field, {1/f}, background, touches, add, false, false, false, sopts);
 		wout = wout.aggregate({agx, agy}, "sum", true, opt);
@@ -229,9 +424,17 @@ SpatRaster SpatRaster::rasterize(SpatVector x, std::string field, std::vector<do
 		}
 	}
 
+	size_t nGeoms = x.size();
+	if (nGeoms == 0) {
+		if (!update) {
+			out = out.init({background}, opt);
+		}
+		return out;
+	}
+
 	if (ispol && touches && add) {
 		add = false;
-		out.addWarning("you cannot use add and touches at the same time");
+		out.addWarning("you cannot use 'sum' and 'touches' at the same time");
 	}
 
 	if (field != "") {
@@ -241,13 +444,23 @@ SpatRaster SpatRaster::rasterize(SpatVector x, std::string field, std::vector<do
 			return out;
 		}
 		std::string dt = x.df.get_datatype(field);
-		if (dt == "string") {
-			//std::vector<std::string> ss = ;
-			SpatFactor f;
-			f.set_values(x.df.getS(i));
+		if (dt == "double") {
+			values = x.df.getD(i);
+		} else if (dt == "long") {
+			values = x.df.as_double(i);
+			out.setValueType(1);
+		} else if (dt == "bool") {
+			values = x.df.as_double(i);
+			out.setValueType(3);
+		} else if (dt == "time") {
+			// tbd
+			values = x.df.as_double(i);
+		} else {
+			std::vector<std::string> sv = x.df.as_string(i);
+			SpatFactor f(sv);
 			values.resize(f.v.size());
-			for (size_t i=0; i<values.size(); i++) {
-				values[i] = f.v[i];
+			for (size_t j=0; j<values.size(); j++) {
+				values[j] = f.v[j];
 			}
 			if (!add && !update) {
 				std::vector<long> u(f.labels.size());
@@ -259,17 +472,9 @@ SpatRaster SpatRaster::rasterize(SpatVector x, std::string field, std::vector<do
 				add = false;
 				addWarning("cannot add factors");
 			}
-		} else if (dt == "double") {
-			values = x.df.getD(i);
-		} else {
-			std::vector<long> v = x.df.getI(i);
-			values.resize(v.size());
-			for (size_t i=0; i<values.size(); i++) {
-				values[i] = v[i];
-			}
 		}
 	}
-	size_t nGeoms = x.size();
+
 	if (values.size() != nGeoms) {
 		recycle(values, nGeoms);
 	}
@@ -281,10 +486,10 @@ SpatRaster SpatRaster::rasterize(SpatVector x, std::string field, std::vector<do
 	}
     std::vector<OGRGeometry *> ogrGeoms;
 	ogrGeoms.reserve(nGeoms);
-	
+
 	OGRLayer *poLayer = vecDS->GetLayer(0);
 	poLayer->ResetReading();
-	
+
 	OGRFeature *poFeature;
 	while( (poFeature = poLayer->GetNextFeature()) != NULL ) {
 		OGRGeometry *poGeometry = poFeature->GetGeometryRef();
@@ -315,10 +520,10 @@ SpatRaster SpatRaster::rasterize(SpatVector x, std::string field, std::vector<do
 	CPLErr err;
 	if (ispol && touches && (nGeoms > 1)) {
 		// first to get the touches
-		papszOptions = CSLSetNameValue(papszOptions, "ALL_TOUCHED", "TRUE"); 
-		err = GDALRasterizeGeometries(rstDS, 
+		papszOptions = CSLSetNameValue(papszOptions, "ALL_TOUCHED", "TRUE");
+		err = GDALRasterizeGeometries(rstDS,
 				static_cast<int>(bands.size()), &(bands[0]),
-				static_cast<int>(ogrGeoms.size()), 
+				static_cast<int>(ogrGeoms.size()),
 				(OGRGeometryH *) &(ogrGeoms[0]),
 				NULL, NULL, &(values[0]), papszOptions, NULL, NULL);
 		CSLDestroy(papszOptions);
@@ -333,21 +538,21 @@ SpatRaster SpatRaster::rasterize(SpatVector x, std::string field, std::vector<do
 		}
 		//GDALFlushCache(rstDS);
 		// second time to fix the internal area
-		err = GDALRasterizeGeometries(rstDS, 
+		err = GDALRasterizeGeometries(rstDS,
 				static_cast<int>(bands.size()), &(bands[0]),
-				static_cast<int>(ogrGeoms.size()), 
+				static_cast<int>(ogrGeoms.size()),
 				(OGRGeometryH *) &(ogrGeoms[0]),
 				NULL, NULL, &(values[0]), NULL, NULL, NULL);
 
 	} else {
 		if (touches) {
-			papszOptions = CSLSetNameValue(papszOptions, "ALL_TOUCHED", "TRUE"); 
+			papszOptions = CSLSetNameValue(papszOptions, "ALL_TOUCHED", "TRUE");
 		} else if (add) {
-			papszOptions = CSLSetNameValue(papszOptions, "MERGE_ALG", "ADD"); 
+			papszOptions = CSLSetNameValue(papszOptions, "MERGE_ALG", "ADD");
 		}
-		err = GDALRasterizeGeometries(rstDS, 
+		err = GDALRasterizeGeometries(rstDS,
 				static_cast<int>(bands.size()), &(bands[0]),
-				static_cast<int>(ogrGeoms.size()), 
+				static_cast<int>(ogrGeoms.size()),
 				(OGRGeometryH *) &(ogrGeoms[0]),
 				NULL, NULL, &(values[0]), papszOptions, NULL, NULL);
 
@@ -380,7 +585,7 @@ SpatRaster SpatRaster::rasterize(SpatVector x, std::string field, std::vector<do
 
 	GDALClose(rstDS);
 	if (driver != "MEM") {
-		out = SpatRaster(filename, {-1}, {""}, {});
+		out = SpatRaster(filename, {-1}, {""}, {}, {});
 	} else {
 		std::string fname = opt.get_filename();
 		if ((fname != "") && (!update)) {
@@ -391,28 +596,43 @@ SpatRaster SpatRaster::rasterize(SpatVector x, std::string field, std::vector<do
 }
 
 
-std::vector<double> SpatRaster::rasterizeCells(SpatVector &v, bool touches, SpatOptions &opt) { 
+std::vector<double> SpatRaster::rasterizeCells(SpatVector &v, bool touches, SpatOptions &opt) {
 // note that this is only for lines and polygons
     SpatOptions ropt(opt);
 	SpatRaster r = geometry(1);
 	SpatExtent e = getExtent();
-	e = e.intersect(v.getExtent());
+	SpatExtent ev = v.getExtent();
+	if (ev.xmin >= ev.xmax) {
+		double xr = 0.1 * xres();
+		ev.xmin -= xr;
+		ev.xmax += xr;
+	}
+	if (ev.ymin >= ev.ymax) {
+		double yr = 0.1 * yres();
+		ev.ymin -= yr;
+		ev.ymax += yr;
+	}
+
+	e = e.intersect(ev);
 	if ( !e.valid() ) {
 		std::vector<double> out(1, NAN);
 		return out;
 	}
 
-	SpatRaster rc = r.crop(e, "out", ropt);
+	SpatRaster rc = r.crop(e, "out", false, ropt);
 	std::vector<double> feats(1, 1) ;
-    SpatRaster rcr = rc.rasterize(v, "", feats, NAN, touches, false, false, false, false, ropt); 
-	SpatVector pts = rcr.as_points(false, true, ropt);
+    SpatRaster rcr = rc.rasterize(v, "", feats, NAN, touches, false, false, false, false, ropt);
+	SpatVector pts = rcr.as_points(false, true, false, ropt);
 	std::vector<double> cells;
 	if (pts.size() == 0) {
 		pts = v.as_points(false, true);
 		SpatDataFrame vd = pts.getGeometryDF();
 		std::vector<double> x = vd.getD(0);
 		std::vector<double> y = vd.getD(1);
-		cells = r.cellFromXY(x, y);
+		cells = r.cellFromXY(x, y);	
+		cells.erase(std::remove_if(cells.begin(), cells.end(),
+                    [](const double& value) { return std::isnan(value); }), cells.end());
+		std::sort( cells.begin(), cells.end() );
 		cells.erase(std::unique(cells.begin(), cells.end()), cells.end());
 		if (cells.size() == 0) {
 			cells.resize(1, NAN);
@@ -422,56 +642,66 @@ std::vector<double> SpatRaster::rasterizeCells(SpatVector &v, bool touches, Spat
 		std::vector<double> x = vd.getD(0);
 		std::vector<double> y = vd.getD(1);
 		cells = r.cellFromXY(x, y);
+//		cells.erase(std::unique(cells.begin(), cells.end()), cells.end());
+		if (cells.size() == 0) {
+			cells.resize(1, NAN);
+		}
 	}
 	return cells;
 }
 
-void SpatRaster::rasterizeCellsWeights(std::vector<double> &cells, std::vector<double> &weights, SpatVector &v, SpatOptions &opt) { 
+void SpatRaster::rasterizeCellsWeights(std::vector<double> &cells, std::vector<double> &weights, SpatVector &v, SpatOptions &opt) {
 // note that this is only for polygons
     SpatOptions ropt(opt);
-	opt.progress = nrow()+1;
-	SpatRaster rr = geometry(1);
-	std::vector<unsigned> fact = {10, 10};
+	//opt.progress = nrow()+1;
+	SpatRaster r = geometry(1);
+	//std::vector<unsigned> fact = {10, 10};
 	SpatExtent e = getExtent();
 	SpatExtent ve = v.getExtent();
 	e = e.intersect(ve);
 	if ( !e.valid() ) {
 		return;
 	}
-	SpatRaster r = rr.crop(v.extent, "out", ropt);
-	r = r.disaggregate(fact, ropt);
-	std::vector<double> feats(1, 1) ;
-	r = r.rasterize(v, "", feats, NAN, true, false, false, false, false, ropt); 
-	r = r.arith(100.0, "/", false, ropt);
-	r = r.aggregate(fact, "sum", true, ropt);
-	SpatVector pts = r.as_points(true, true, ropt);
-	if (pts.size() == 0) {
+	bool cropped = false;
+	SpatRaster rc = r.crop(v.extent, "out", false, ropt);
+	if ( ((ncol() > 1000) && ((ncol() / rc.ncol()) > 1.5))
+			|| ((nrow() > 1000) && ((nrow() / rc.nrow()) > 1.5) )) {
+		cropped = true;
+		r = rc;
+	}
+	std::vector<double> feats;
+	r = r.rasterize(v, "", feats, NAN, false, false, true, false, false, ropt);
+	std::vector<std::vector<double>> cv = r.cells_notna(ropt);
+
+	if (cv[0].size() == 0) {
 		weights.resize(1);
 		weights[0] = NAN;
 		cells.resize(1);
 		cells[0] = NAN;
 	} else {
-		SpatDataFrame vd = pts.getGeometryDF();
-		std::vector<double> x = vd.getD(0);
-		std::vector<double> y = vd.getD(1);
-		cells = rr.cellFromXY(x, y);
-		weights = pts.df.dv[0];
+		weights = cv[1];
+		if (cropped) {
+			cv = r.xyFromCell(cv[0]);
+			cells = cellFromXY(cv[0], cv[1]);
+		} else {
+			cells = cv[0];
+		}
 	}
 	return;
 }
 
-void SpatRaster::rasterizeCellsExact(std::vector<double> &cells, std::vector<double> &weights, SpatVector &v, SpatOptions &opt) { 
-    
+void SpatRaster::rasterizeCellsExact(std::vector<double> &cells, std::vector<double> &weights, SpatVector &v, SpatOptions &opt) {
+
 	SpatOptions ropt(opt);
 	opt.progress = nrow()+1;
 	SpatRaster r = geometry(1);
-	r = r.crop(v.extent, "out", ropt);
+	r = r.crop(v.extent, "out", false, ropt);
 
 //	if (r.ncell() < 1000) {
 		std::vector<double> feats(1, 1) ;
-		r = r.rasterize(v, "", feats, NAN, true, false, false, false, false, ropt); 
+		r = r.rasterize(v, "", feats, NAN, true, false, false, false, false, ropt);
 
-		SpatVector pts = r.as_points(true, true, ropt);
+		SpatVector pts = r.as_points(true, true, false, ropt);
 		if (pts.size() == 0) {
 			weights.resize(1);
 			weights[0] = NAN;
@@ -483,7 +713,7 @@ void SpatRaster::rasterizeCellsExact(std::vector<double> &cells, std::vector<dou
 			std::vector<double> y = vd.getD(1);
 			cells = cellFromXY(x, y);
 
-			SpatVector rv = r.as_polygons(false, false, false, true, ropt);
+			SpatVector rv = r.as_polygons(false, false, false, true, false, ropt);
 			std::vector<double> csize = rv.area("m", true, {});
 			rv.df.add_column(csize, "area");
 			rv.df.add_column(cells, "cells");
@@ -503,7 +733,7 @@ void SpatRaster::rasterizeCellsExact(std::vector<double> &cells, std::vector<dou
 	else {
 		std::vector<double> feats(1, 1) ;
 		SpatVector vv = v.as_lines();
-		SpatRaster b = r.rasterize(vv, "", feats, NAN, true, false, false, false, false, opt); 
+		SpatRaster b = r.rasterize(vv, "", feats, NAN, true, false, false, false, false, opt);
 		SpatVector pts = b.as_points(true, true, opt);
 		if (pts.nrow() > 0) {
 			SpatDataFrame vd = pts.getGeometryDF();
@@ -523,7 +753,7 @@ void SpatRaster::rasterizeCellsExact(std::vector<double> &cells, std::vector<dou
 			cells = bv.df.dv[1];
 		}
 		// touches = false
-		r = r.rasterize(v, "", feats, NAN, false, false, false, false, false, opt); 
+		r = r.rasterize(v, "", feats, NAN, false, false, false, false, false, opt);
 		pts = r.as_points(true, true, opt);
 		if (pts.nrow() > 0) {
 			SpatDataFrame vd = pts.getGeometryDF();
@@ -546,4 +776,51 @@ void SpatRaster::rasterizeCellsExact(std::vector<double> &cells, std::vector<dou
 }
 
 
+void SpatRaster::rasterizeLinesLength(std::vector<double> &cells, std::vector<double> &weights, SpatVector &v, SpatOptions &opt) {
+
+	if (v.type() != "lines") {
+		setError("expected lines");
+		return;
+	}
+
+/*
+	double m = 1;
+	if (!v.is_lonlat()) {
+		double tom = v.srs.to_meter();
+		tom = std::isnan(tom) ? 1 : tom;
+		m *= tom;
+	}
+*/
+
+	SpatOptions xopt(opt);
+	xopt.ncopies = std::max(xopt.ncopies, (unsigned)4) * 8;
+	SpatRaster x = geometry(1);
+
+	SpatExtent ev = v.getExtent();
+	x = x.crop(ev, "out", false, xopt);
+	BlockSize bs = x.getBlockSize(xopt);
+
+	SpatExtent e = x.getExtent();
+	double rsy = x.yres() / 2;
+	for (size_t i=0; i < bs.n; i++) {
+		e.ymax = yFromRow(bs.row[i]) + rsy;
+		e.ymin = yFromRow(bs.row[i] + bs.nrows[i] - 1) - rsy;
+		SpatRaster tmp = x.crop(e, "near", false, xopt);
+		std::vector<double> cell(tmp.ncell());
+		std::iota(cell.begin(), cell.end(), 0);
+		std::vector<std::vector<double>> xy = tmp.xyFromCell(cell);
+		cell = cellFromXY(xy[0], xy[1]);
+		SpatVector p = tmp.as_polygons(true, false, false, false, false, xopt);
+		p.df.add_column(cell, "cell");
+		p = p.intersect(v, true);
+		if (p.nrow() > 1) {
+			cells.insert(cells.end(), p.df.dv[0].begin(), p.df.dv[0].end());
+			std::vector<double> w = p.length();
+			double sm = std::accumulate(w.begin(), w.end(), 0.0);
+			for (double &d : w) d /= sm;
+			weights.insert(weights.end(), w.begin(), w.end());
+		}
+	}
+}
+
 
