@@ -1,4 +1,4 @@
-// Copyright (c) 2018-2022  Robert J. Hijmans
+// Copyright (c) 2018-2023  Robert J. Hijmans
 //
 // This file is part of the "spat" library.
 //
@@ -155,20 +155,21 @@ SpatCategories GetRAT(GDALRasterAttributeTable *pRAT) {
 	size_t nc = (int) pRAT->GetColumnCount();
 	size_t nr = (int) pRAT->GetRowCount();
 
-
 	std::vector<std::string> ss = {"histogram", "count", "red", "green", "blue", "opacity", "r", "g", "b", "alpha"};
 
-	std::vector<std::string> ratnms;
+	//std::vector<std::string> ratnms;
 	std::vector<int> id, id2;
 
 	bool hasvalue=false;
+	//std::string valuename = "";
 	for (size_t i=0; i<nc; i++) {
 		std::string name = pRAT->GetNameOfCol(i);
 		lowercase(name);
-		ratnms.push_back(name);
-		if (name == "value") {
+		//ratnms.push_back(name);
+		if (!hasvalue && ((name == "value") || (name == "id"))) {
 			id.insert(id.begin(), i);
 			hasvalue = true;
+			//valuename = name;
 		} else {
 			int k = where_in_vector(name, ss, true);
 			if (k >= 0) {
@@ -178,6 +179,10 @@ SpatCategories GetRAT(GDALRasterAttributeTable *pRAT) {
 			}
 		}
 	}
+	if ((id.size() <= 1) && (id2.size() == 1)) {
+// #790 avoid having just "count" or "histogram" 
+		return(out);
+	}
 	id.insert(id.end(), id2.begin(), id2.end());
 
 	if (!hasvalue) {
@@ -230,7 +235,7 @@ bool GetVAT(std::string filename, SpatCategories &vat) {
 	SpatVector v, fvct;
 	std::vector<double> fext;
 
-	v.read(filename, "", "", fext, fvct, false);
+	v.read(filename, "", "", fext, fvct, false, ""); 
 	if (v.df.nrow() == 0) return false;
 
 
@@ -333,6 +338,7 @@ bool setIntCol(SpatDataFrame &d, SpatDataFrame &out, int k, std::string name) {
 
 bool colsFromRat(SpatDataFrame &d, SpatDataFrame &out) {
 
+
 	std::vector<std::string> ss = d.get_names();
 	for (size_t i=0; i<ss.size(); i++) {
 		lowercase(ss[i]);
@@ -341,9 +347,10 @@ bool colsFromRat(SpatDataFrame &d, SpatDataFrame &out) {
 	int k = where_in_vector("value", ss, true);
 	if (k >= 0) {
 		size_t j = d.iplace[k];
+		
 		if (d.itype[k] == 1) {
 			out.add_column(d.iv[j], "value");
-		} else {
+		} else if (d.itype[k] == 0) {
 			std::vector<long> x;
 			x.reserve(d.nrow());
 			for (size_t i=0; i<d.nrow(); i++) {
@@ -355,39 +362,48 @@ bool colsFromRat(SpatDataFrame &d, SpatDataFrame &out) {
 		return false;
 	}
 
-	std::vector<std::string> cols1 = {"red", "green", "blue"};
-	std::vector<std::string> cols2 = {"r", "g", "b"};
-	for (size_t i=0; i<3; i++) {
-		int k = where_in_vector(cols1[i], ss, true);
-		if (k >= 0) {
-			if (!setIntCol(d, out, k, cols1[i])) return false;
-		} else {
-			int k = where_in_vector(cols2[i], ss, true);
+
+//	int k = where_in_vector("colors", ss, true);
+//	if (k >= 0) {
+		//col2rgb(d, out, k);
+//	} else {
+		std::vector<std::string> cols1 = {"red", "green", "blue"};
+		std::vector<std::string> cols2 = {"r", "g", "b"};
+		for (size_t i=0; i<3; i++) {
+			int k = where_in_vector(cols1[i], ss, true);
 			if (k >= 0) {
 				if (!setIntCol(d, out, k, cols1[i])) return false;
 			} else {
-				return false;
+				int k = where_in_vector(cols2[i], ss, true);
+				if (k >= 0) {
+					if (!setIntCol(d, out, k, cols1[i])) return false;
+				} else {
+					return false;
+				}
 			}
 		}
-	}
 
-	bool have_alpha = false;
-	k = where_in_vector("alpha", ss, true);
-	if (k >= 0) {
-		if (setIntCol(d, out, k, "alpha")) have_alpha = true;
-	}
-	if (!have_alpha) {
-		int k = where_in_vector("transparency", ss, true);
+		bool have_alpha = false;
+		k = where_in_vector("alpha", ss, true);
 		if (k >= 0) {
 			if (setIntCol(d, out, k, "alpha")) have_alpha = true;
 		}
-	}
-	if (!have_alpha) {
-		std::vector<long> a(out.nrow(), 255);
-		out.add_column(a, "alpha");
-	}
 
+		if (!have_alpha) {
+			int k = where_in_vector("transparency", ss, true);
+			if (k >= 0) {
+				if (setIntCol(d, out, k, "alpha")) have_alpha = true;
+			}
+		}
+
+		if (!have_alpha) {
+			std::vector<long> a(out.nrow(), 255);
+			out.add_column(a, "alpha");
+		}
+//	}
+	
 	d = out;
+
 	return true;
 }
 
@@ -455,6 +471,7 @@ std::string basename_sds(std::string f) {
 	if (std::string::npos != i) {
 		f.erase(0, i + 1);
 	}
+	// this may be incorrect of the variable name includes a ":" ?
 	const size_t j = f.find_last_of(":");
 	if (std::string::npos != j) {
 		f.erase(0, j + 1);
@@ -535,6 +552,21 @@ std::string getDsPRJ(GDALDataset *poDataset) {
 }
 
 
+inline std::string dtypename(const std::string &d) {
+	if (d == "Float64") return "FLT8S";
+	if (d == "Float32") return "FLT4S";
+	if (d == "Int64") return "INT8S";
+	if (d == "Int32") return "INT4S";
+	if (d == "Int16") return "INT2S";
+	if (d == "Int8") return "INT1S";
+	if (d == "UInt64") return "INT8U";
+	if (d == "UInt32") return "INT4U";
+	if (d == "UInt16") return "INT2U";
+	if (d == "Byte") return "INT1U";
+	return "FLT4S";
+}
+
+
 
 SpatRasterStack::SpatRasterStack(std::string fname, std::vector<int> ids, bool useids) {
 
@@ -567,7 +599,6 @@ SpatRasterStack::SpatRasterStack(std::string fname, std::vector<int> ids, bool u
 		ids.resize(meta.size());
 		std::iota(ids.begin(), ids.end(), 0);
 	}
-	SpatRaster sub;
 	int idssz = ids.size();
 	int metsz = meta.size();
 
@@ -582,8 +613,10 @@ SpatRasterStack::SpatRasterStack(std::string fname, std::vector<int> ids, bool u
 			size_t pos = s.find(delim);
 			if (pos != std::string::npos) {
 				s.erase(0, pos + delim.length());
+				SpatRaster sub;
 				if (sub.constructFromFile(s, {-1}, {""}, {}, {})) {
-					if (!push_back(sub, basename_sds(s), "", "", true)) {
+					std::string sname = (sub.source[0].source_name != "") ? sub.source[0].source_name : basename_sds(s);
+					if (!push_back(sub, sname, sub.source[0].source_name_long, sub.source[0].unit[0], true)) {
 						addWarning("skipped (different geometry): " + s);
 					}
 				} else {
@@ -596,6 +629,64 @@ SpatRasterStack::SpatRasterStack(std::string fname, std::vector<int> ids, bool u
 }
 
 
+SpatRasterCollection::SpatRasterCollection(std::string fname, std::vector<int> ids, bool useids) {
+
+	std::vector<std::string> ops;
+    GDALDataset *poDataset = openGDAL(fname, GDAL_OF_RASTER | GDAL_OF_READONLY | GDAL_OF_VERBOSE_ERROR, ops, ops);
+    if( poDataset == NULL )  {
+		if (!file_exists(fname)) {
+			setError("file does not exist");
+		} else {
+			setError("cannot read from " + fname );
+		}
+		return;
+	}
+
+	std::string delim = "NAME=";
+	char **metadata = poDataset->GetMetadata("SUBDATASETS");
+
+	if (metadata == NULL) {
+		setError("file has no subdatasets");
+		GDALClose( (GDALDatasetH) poDataset );
+		return;
+	}
+
+	std::vector<std::string> meta;
+    for (size_t i=0; metadata[i] != NULL; i++) {
+		meta.push_back(metadata[i]);
+	}
+
+	if (!useids) {
+		ids.resize(meta.size());
+		std::iota(ids.begin(), ids.end(), 0);
+	}
+	int idssz = ids.size();
+	int metsz = meta.size();
+
+	if (metsz == 0) {
+		setError("file does not consist of subdatasets");
+	} else {
+		for (int i=0; i<idssz; i++) {
+			if ((ids[i] < 0) || ((2*ids[i]) >= metsz)) {
+				continue;
+			}
+			std::string s = meta[ids[i]*2];
+			size_t pos = s.find(delim);
+			if (pos != std::string::npos) {
+				s.erase(0, pos + delim.length());
+				SpatRaster sub;
+				if (sub.constructFromFile(s, {-1}, {""}, {}, {})) {
+					push_back(sub, basename_sds(s));
+				} else {
+					addWarning("skipped (fail): " + s);
+				}
+			}
+		}
+	}
+	GDALClose( (GDALDatasetH) poDataset );
+}
+
+
 /*
 SpatRaster SpatRaster::fromFiles(std::vector<std::string> fname, std::vector<int> subds, std::vector<std::string> subdsname, std::vector<std::string> drivers, std::vector<std::string> options) {
 	SpatRaster out;
@@ -634,7 +725,7 @@ bool SpatRaster::constructFromFile(std::string fname, std::vector<int> subds, st
 		if (!file_exists(fname)) {
 			setError("file does not exist: " + fname);
 		} else {
-			setError("cannot read from " + fname );
+			setError("cannot open this file as a SpatRaster: " + fname);
 		}
 		return false;
 	}
@@ -692,7 +783,7 @@ bool SpatRaster::constructFromFile(std::string fname, std::vector<int> subds, st
 		}
 	} else {
 		hasExtent = false;
-		SpatExtent e(0, 1, 0, 1);
+		SpatExtent e(0, s.ncol, 0, s.nrow);
 		s.extent = e;
 		if ((gdrv=="netCDF") || (gdrv == "HDF5")) {
 			#ifndef standalone
@@ -807,8 +898,7 @@ bool SpatRaster::constructFromFile(std::string fname, std::vector<int> subds, st
 		poBand->GetBlockSize(&bs1, &bs2);
 		s.blockcols[i] = bs1;
 		s.blockrows[i] = bs2;
-
-		//std::string dtype = GDALGetDataTypeName(poBand->GetRasterDataType());
+		s.dtype = dtypename(GDALGetDataTypeName(poBand->GetRasterDataType()));
 
 		adfMinMax[0] = poBand->GetMinimum( &bGotMin );
 		adfMinMax[1] = poBand->GetMaximum( &bGotMax );
@@ -894,6 +984,7 @@ bool SpatRaster::constructFromFile(std::string fname, std::vector<int> subds, st
 				nm = nms[s.cats[i].index];
 			}
 		}
+		
 		if (nm == "") {
 			if (bandname != "") {
 				nm = bandname;
@@ -916,8 +1007,10 @@ bool SpatRaster::constructFromFile(std::string fname, std::vector<int> subds, st
 	if ((gdrv=="netCDF") || (gdrv == "HDF5"))  {
 		std::vector<std::string> metadata;
 		char **m = poDataset->GetMetadata();
-		while (*m != nullptr) {
-			metadata.push_back(*m++);
+		if (m) {
+			while (*m != nullptr) {
+				metadata.push_back(*m++);
+			}
 		}
 		s.set_names_time_ncdf(metadata, bandmeta, msg);
 	} else if (gdrv == "GRIB") {
@@ -981,6 +1074,7 @@ void NAso(std::vector<double> &d, size_t n, const std::vector<double> &flags, co
 				std::replace(d.begin()+start, d.begin()+start+n, flag, na);
 			}
 		}
+		
 		if (haveso[i]) {
 			for (size_t j=start; j<(start+n); j++) {
 				d[j] = d[j] * scale[i] + offset[i];
@@ -1010,12 +1104,10 @@ void vflip(std::vector<double> &v, const size_t &ncell, const size_t &nrows, con
 
 void SpatRaster::readChunkGDAL(std::vector<double> &data, unsigned src, size_t row, unsigned nrows, size_t col, unsigned ncols) {
 
-
 	if (source[src].flipped) {
 		row = nrow() - row - nrows;
 	}
 
-
 	if (source[src].multidim) {
 		readValuesMulti(data, src, row, nrows, col, ncols);
 		return;
@@ -1672,6 +1764,7 @@ std::vector<int_64> ncdf_time(const std::vector<std::string> &metadata, std::vec
 		if (fc & fu) break;
 	}
 
+	bool years = false;
 	bool days = false;
 	bool hours = false;
 	bool seconds = false;
@@ -1684,6 +1777,8 @@ std::vector<int_64> ncdf_time(const std::vector<std::string> &metadata, std::vec
 			days = true;
 		} else if ((origin.find("seconds")) != std::string::npos) {
 			seconds = true;
+		} else if ((origin.find("years")) != std::string::npos) {
+			years = true;
 		}
 		size_t pos;
 		if ((pos = origin.find("from")) != std::string::npos) {
@@ -1720,6 +1815,10 @@ std::vector<int_64> ncdf_time(const std::vector<std::string> &metadata, std::vec
 		} else if (seconds) {
 			offset = get_time_string(origin);
 			for (size_t i=0; i<raw.size(); i++) out.push_back(raw[i]+offset);
+		} else if (years) {
+			step = "years";
+			int syear = getyear(origin);
+			for (size_t i=0; i<raw.size(); i++) out.push_back(get_time(syear+raw[i], 6, 30, 0, 0, 0));
 		} else {
 			step = "raw";
 			for (size_t i=0; i<raw.size(); i++) out.push_back(raw[i]);
@@ -1823,12 +1922,18 @@ void SpatRasterSource::set_names_time_ncdf(std::vector<std::string> metadata, st
 
 std::vector<std::vector<std::string>> grib_names(const std::vector<std::vector<std::string>> &m) {
 
-	std::vector<std::vector<std::string>> out(3);
+	std::vector<std::vector<std::string>> out(4);
 	if (m.size() < 1) return out;
+	
+	bool ft1 = false;
+	bool ft2 = false;
 
 	for (size_t i=0; i<m.size(); i++) {
-		std::string comm, time, units = "";
+
+		std::string comm, time1, time2, units = "";
+
 		for (size_t j=0; j<m[i].size(); j++) {
+			
 			size_t pos = m[i][j].find("GRIB_COMMENT=");
 			if (pos != std::string::npos) {
 				comm = m[i][j];
@@ -1845,6 +1950,18 @@ std::vector<std::vector<std::string>> grib_names(const std::vector<std::vector<s
 				lrtrim(units);
 				continue;
 			}
+			pos = m[i][j].find("GRIB_VALID_TIME=");
+			if (pos != std::string::npos) {
+				std::string tmp = m[i][j];
+				tmp.erase(0, pos+16);
+				pos = tmp.find("sec");
+				if (pos != std::string::npos) {
+					tmp.erase(tmp.begin()+pos, tmp.end());
+				}
+				time1 = tmp;
+				ft1 = true;
+				continue;
+			}
 			pos = m[i][j].find("TIME=");
 			if (pos != std::string::npos) {
 				std::string tmp = m[i][j];
@@ -1852,15 +1969,23 @@ std::vector<std::vector<std::string>> grib_names(const std::vector<std::vector<s
 				pos = tmp.find("sec");
 				if (pos != std::string::npos) {
 					tmp.erase(tmp.begin()+pos, tmp.end());
-					time = tmp;
 				}
-				continue;
+				time2 = tmp;
+				ft2 = true;
 			}
 		}
 		out[0].push_back(comm);
 		out[1].push_back(units);
-		out[2].push_back(time);
+		out[2].push_back(time1);
+		out[3].push_back(time2);
 	}
+	
+	if (!ft1) {
+		if (ft2) {
+			out[2] = out[3];
+		}
+	}
+	out.resize(3);
 	return out;
 }
 
@@ -1869,6 +1994,7 @@ std::vector<std::vector<std::string>> grib_names(const std::vector<std::vector<s
 void SpatRasterSource::set_names_time_grib(std::vector<std::vector<std::string>> bandmeta, std::string &msg) {
 
 	if (bandmeta.size() == 0) return;
+	
 	std::vector<std::vector<std::string>> nms = grib_names(bandmeta);
 
 	if (nms[0].size() != names.size()) return;
@@ -1878,28 +2004,36 @@ void SpatRasterSource::set_names_time_grib(std::vector<std::vector<std::string>>
 		str_replace(names[i], "0[-] ", "");
 		str_replace_all(names[i], "\"", "");
 	}
-	unit = {nms[1]};
 
-	std::vector<int_64> tm;
-	bool hastime = true;
-	for (size_t i=0; i<nms[2].size(); i++) {
-		if (nms[2][i] == "") {
-			hastime = false;
-			break;
-		}
-		int_64 tim;
-		try {
-			tim = stol(nms[2][i]);
-		} catch(...) {
-			hastime = false;
-			break;
-		}
-		tm.push_back(tim);
+	if (nms[1].size() == nms[0].size()) {
+		unit = {nms[1]};
 	}
+
+	bool hastime = false;
+	std::vector<int_64> tm;
+	if (nms[2].size() == nms[0].size()) {
+		hastime = true;
+		int_64 tim;
+		for (size_t i=0; i<nms[2].size(); i++) {
+			if (nms[2][i] == "") {
+				hastime = false;
+				break;
+			}
+			try {
+				tim = stol(nms[2][i]);
+			} catch(...) {
+				hastime = false;
+				break;
+			}
+			tm.push_back(tim);
+		}
+	}
+
 	if (hastime) {
 		time = tm;
 		timestep = "seconds";
 		hasTime = true;
 	}
+	
 }
 
