@@ -1,4 +1,4 @@
-// Copyright (c) 2018-2021  Robert J. Hijmans
+// Copyright (c) 2018-2023  Robert J. Hijmans
 //
 // This file is part of the "spat" library.
 //
@@ -47,7 +47,7 @@ std::vector<double> SpatRaster::readSample(unsigned src, size_t srows, size_t sc
 
 	unsigned nl = source[src].nlyr;
 	std::vector<size_t> oldcol, oldrow;
-	std::vector<double>	out; 
+	std::vector<double>	out;
 	getSampleRowCol(oldrow, oldcol, nrow(), ncol(), srows, scols);
 
 	out.reserve(srows*scols);
@@ -99,6 +99,11 @@ SpatRaster SpatRaster::sampleRegularRaster(unsigned size) {
 	out.source[0].nrow = nr;
 	out.source[0].ncol = nc;
 
+	std::vector<int> vt = getValueType(true);
+	if (vt.size() == 1) {
+		out.setValueType(vt[0]);
+	}
+
 	if (!source[0].hasValues) return (out);
 
 	std::vector<double> v;
@@ -138,6 +143,11 @@ SpatRaster SpatRaster::sampleRowColRaster(size_t nr, size_t nc) {
 	out.source[0].nrow = nr;
 	out.source[0].ncol = nc;
 
+	std::vector<int> vt = getValueType(true);
+	if (vt.size() == 1) {
+		out.setValueType(vt[0]);
+	}
+
 	if (!source[0].hasValues) return (out);
 
 	std::vector<double> v;
@@ -255,7 +265,7 @@ std::vector<std::vector<double>> SpatRaster::sampleRowColValues(size_t nr, size_
 
 
 std::vector<size_t> sample_replace(size_t size, size_t N, unsigned seed){
-	std::default_random_engine gen(seed);   
+	std::default_random_engine gen(seed);
 	std::uniform_int_distribution<> U(0, N-1);
 	std::vector<size_t> sample;
 	sample.reserve(size);
@@ -273,7 +283,7 @@ std::vector<size_t> sample_replace_weights(size_t size, size_t N, std::vector<do
 	for (double& d : prob)  d /= maxw;
 	double minw = *min_element(prob.begin(), prob.end());
 
-	std::default_random_engine gen(seed);   
+	std::default_random_engine gen(seed);
 	std::uniform_int_distribution<> U(0, N-1);
 	std::vector<size_t> sample;
 	sample.reserve(size);
@@ -285,7 +295,7 @@ std::vector<size_t> sample_replace_weights(size_t size, size_t N, std::vector<do
 	while (cnt < size) {
 		double w = Uw(gen);
 		double v = U(gen);
-		if (prob[v] >= w) { 
+		if (prob[v] >= w) {
 			sample.push_back(v);
 			cnt++;
 		} else {
@@ -313,7 +323,7 @@ std::vector<size_t> sample_replace_weights_gen(size_t size, size_t N, std::vecto
 	while (cnt < size) {
 		double w = Uw(gen);
 		double v = U(gen);
-		if (prob[v] >= w) { 
+		if (prob[v] >= w) {
 			sample.push_back(v);
 			cnt++;
 		}
@@ -324,14 +334,14 @@ std::vector<size_t> sample_replace_weights_gen(size_t size, size_t N, std::vecto
 
 std::vector<size_t> sample_no_replace(size_t size, size_t N, unsigned seed){
 	size_t one = 1;
-	size = std::max(one, std::min(size, N)); 
+	size = std::max(one, std::min(size, N));
 	std::vector<size_t> sample;
 	if (size == N) {
 		sample.resize(size);
 		std::iota(sample.begin(), sample.end(), 0);
 		return sample;
 	}
-	std::default_random_engine gen(seed);   
+	std::default_random_engine gen(seed);
 
 	if (size >= .66 * N) {
 		sample.resize(N);
@@ -345,7 +355,7 @@ std::vector<size_t> sample_no_replace(size_t size, size_t N, unsigned seed){
 
 	std::uniform_real_distribution<> U( 0, std::nextafter(1.0, std::numeric_limits<double>::max() ) );
 
-	sample.reserve(size); 
+	sample.reserve(size);
 	for (size_t i=0; i<N; i++) {
 		if ( ((N-i) * U(gen)) <  (size - sample.size()) ) {
 			sample.push_back(i);
@@ -360,7 +370,7 @@ std::vector<size_t> sample_no_replace(size_t size, size_t N, unsigned seed){
 
 std::vector<size_t> sample_no_replace_weights(size_t size, size_t N, std::vector<double> prob, unsigned seed){
 	size_t one = 1;
-	size = std::max(one, std::min(size, N)); 
+	size = std::max(one, std::min(size, N));
 	std::vector<size_t> sample;
 	std::default_random_engine gen(seed);
 	if (size == N) {
@@ -372,7 +382,7 @@ std::vector<size_t> sample_no_replace_weights(size_t size, size_t N, std::vector
 
 	std::uniform_int_distribution<> U(0, std::numeric_limits<int>::max());
 	std::unordered_set<size_t> sampleset;
-  
+
 	size_t isize = size;
 	if (size > (0.8 * N)) {
 		isize = N - size;
@@ -384,7 +394,7 @@ std::vector<size_t> sample_no_replace_weights(size_t size, size_t N, std::vector
 			std::vector<size_t> s = sample_replace_weights(ssize, N, prob, seed);
 			for (size_t i=0; i<s.size(); i++) {
 				sampleset.insert(s[i]);
-			} 
+			}
 			cnt++;
 			if (cnt > 10) break;
 		}
@@ -411,7 +421,7 @@ std::vector<size_t> sample_no_replace_weights(size_t size, size_t N, std::vector
 			std::vector<size_t> s = sample_replace_weights(ssize, N, prob, seed);
 			for (size_t i=0; i<s.size(); i++) {
 				sampleset.insert(s[i]);
-			} 
+			}
 			cnt++;
 			if (cnt > 10) break;
 		}
@@ -469,7 +479,7 @@ std::vector<std::vector<double>> SpatRaster::sampleRandomValues(unsigned size, b
 
 	std::vector<double> dcells(cells.begin(), cells.end());
 	std::vector<std::vector<double>> d = extractCell(dcells);
-	return d; 
+	return d;
 }
 
 
@@ -509,7 +519,7 @@ std::vector<size_t> SpatExtent::test_sample(size_t size, size_t N, bool replace,
 std::vector<std::vector<double>> SpatExtent::sampleRandom(size_t size, bool lonlat, unsigned seed){
 	std::vector<std::vector<double>> out(2);
 	if (size == 0) return out;
-	std::default_random_engine gen(seed);   
+	std::default_random_engine gen(seed);
 
 
 	if (lonlat) {
@@ -570,14 +580,14 @@ std::vector<std::vector<double>> SpatExtent::sampleRegular(size_t size, bool lon
 	double r2 = ymax - ymin;
 
 	if (lonlat) {
-		double halfy = ymin + (ymax - ymin)/2;
-		double dx = distance_lonlat(xmin, halfy, xmax, halfy);
+		double halfy = ymin + r2/2;
+		// beware that -180 is the same as 180; and that latitude can only go from -90:90 therefore:
+		double dx = distance_lonlat(xmin, halfy, xmin + 1, halfy) * std::min(180.0, r1);
 		double dy = distance_lonlat(0, ymin, 0, ymax);
-		double ratio = dx/dy;
-		double ny = std::max(1.0, sqrt(size / ratio));
-		double nx = std::max(1.0, size / ny);
-		ny = std::round(ny);
-		nx = std::round(nx);
+		double ratio = dy/dx;
+		double n = sqrt(size);
+		double ny = std::round(std::max(1.0, n * ratio));
+		double nx = std::round(std::max(1.0, n / ratio));
 		double x_i = r1 / nx;
 		double y_i = r2 / ny;
 
@@ -598,27 +608,32 @@ std::vector<std::vector<double>> SpatExtent::sampleRegular(size_t size, bool lon
 		for (size_t i=0; i<w.size(); i++) {
 			xi.push_back(x_i / (w[i] * nwsumw));
 		}
-		double halfx = xmin + (xmax - xmin)/2;
-		for (size_t i=0; i<lat.size(); i++) {
-			double start = halfx - 0.5*xi[i];
-			std::vector <double> x;
-			if (start < xmin) {
-				x = { halfx };
-			} else {
-				while (start > xmin) {
-					start -= xi[i];
-				} 
-				x = seq(start + xi[i], xmax, xi[i]);
+		bool global = (xmax - xmin) > 355; // needs refinement
+		if (global) {
+			xmax -= 0.000001;
+			for (size_t i=0; i<lat.size(); i++) {
+				size_t n = std::max(1, (int)(360.0/xi[i]));
+				double step = 360.0 / n;
+				std::vector<double> x = seq(xmin+0.5*step, xmax, step);
+				std::vector<double> y(x.size(), lat[i]);
+				out[0].insert(out[0].end(), x.begin(), x.end());
+				out[1].insert(out[1].end(), y.begin(), y.end());
 			}
-			if (x.size() <= 1) {
-				x = { halfx };
-			}
-			std::vector <double> y(x.size(), lat[i]);
-        
-			out[0].insert(out[0].end(), x.begin(), x.end());
-			out[1].insert(out[1].end(), y.begin(), y.end());
-		}
 
+		} else {
+			double halfx = xmin + (xmax - xmin)/2;
+			for (size_t i=0; i<lat.size(); i++) {
+				std::vector<double> x = seq(halfx, xmax, xi[i]);
+				double start = halfx-xi[i];
+				if (start > xmin) {
+					std::vector <double> x2 = seq(start, xmin, -xi[i]);
+					x.insert(x.end(), x2.begin(), x2.end());
+				}
+				std::vector<double> y(x.size(), lat[i]);
+				out[0].insert(out[0].end(), x.begin(), x.end());
+				out[1].insert(out[1].end(), y.begin(), y.end());
+			}
+		}
 	} else {
 		double ratio = r1/r2;
 		double ny = std::max(1.0, sqrt(size / ratio));
@@ -652,7 +667,7 @@ std::vector<std::vector<double>> SpatExtent::sampleRegular(size_t size, bool lon
 
 std::vector<size_t> SpatRaster::sampleCells(unsigned size, std::string method, bool replace, unsigned seed) {
 
-	std::default_random_engine gen(seed);   
+	std::default_random_engine gen(seed);
 	std::vector<size_t> out;
 	if ((size >= ncell()) & (!replace)) {
 		out.resize(ncell());
@@ -667,7 +682,7 @@ std::vector<size_t> SpatRaster::sampleCells(unsigned size, std::string method, b
 
 	} else if (method == "regular") {
 
-	} else { //method == "stratified" 
+	} else { //method == "stratified"
 
 	} // else "Cluster"
 	return out;
@@ -710,7 +725,7 @@ SpatVector SpatVector::sample(unsigned n, std::string method, unsigned seed) {
 			}
 			SpatVector s = subset_rows(g);
 			s = s.sample(n, "random", false, "", seed);
-			for (long &v : s.df.iv[0]) v = v+i; 
+			for (long &v : s.df.iv[0]) v = v+i;
 			out = out.append(s, true);
 		}
 		return out;
@@ -795,12 +810,12 @@ SpatVector SpatVector::sample(unsigned n, std::string method, unsigned seed) {
 		pxy = extent.sampleRegular(ssize, lonlat);
 	}
 	out = SpatVector(pxy[0], pxy[1], points, "");
-	out = intersect(out);
+	out = intersect(out, true);
 	if (random) {
 		if (out.size() > n) {
 			std::vector<int> rows(out.size());
 			std::iota(rows.begin(), rows.end(), 0);
-			std::default_random_engine gen(seed);   
+			std::default_random_engine gen(seed);
 			std::shuffle(rows.begin(), rows.end(), gen);
 			rows.resize(n);
 			out = out.subset_rows(rows);
@@ -846,7 +861,7 @@ SpatVector SpatVector::sample_geom(std::vector<unsigned> n, std::string method,
 std::vector<double> sample(size_t size, size_t N, bool replace, std::vector<double> prob, unsigned seed){
     // Sample "size" elements from [1, N]
 	std::vector<double> result;
-	std::default_random_engine gen(seed);   
+	std::default_random_engine gen(seed);
 
 	bool weights = false;
 	if (prob.size() == N) {
@@ -867,7 +882,7 @@ std::vector<double> sample(size_t size, size_t N, bool replace, std::vector<doub
 			while (cnt < size) {
 				double w = wdist(gen);
 				double v = distribution(gen);
-				if (prob[v] >= w) { 
+				if (prob[v] >= w) {
 					result.push_back(v);
 					cnt++;
 				}
@@ -875,7 +890,7 @@ std::vector<double> sample(size_t size, size_t N, bool replace, std::vector<doub
 		} else {
 			for (size_t i=0; i<size; i++) {
 				result.push_back( distribution(gen) );
-			} 
+			}
 		}
 	} else {//without replacement
 
@@ -891,7 +906,7 @@ std::vector<double> sample(size_t size, size_t N, bool replace, std::vector<doub
 			while (cnt < size) {
 				double w = wdist(gen)/N;
 				double v = distribution(gen);
-				if (prob[v] >= w) { 
+				if (prob[v] >= w) {
 					if (!samples.insert(v).second) {
 						samples.insert(r);
 						cnt++;
@@ -904,10 +919,10 @@ std::vector<double> sample(size_t size, size_t N, bool replace, std::vector<doub
 			for (size_t r = N - size; r < N; ++r) {
 				unsigned v = distribution(gen) - 1;
 				if (!samples.insert(v).second) samples.insert(r);
-			} 
+			}
 		}
 		result = std::vector<double>(samples.begin(), samples.end());
-		std::shuffle(result.begin(), result.end(), gen);    
+		std::shuffle(result.begin(), result.end(), gen);
 	}
     return result;
 }
