@@ -1,4 +1,4 @@
-// Copyright (c) 2018-2021  Robert J. Hijmans
+// Copyright (c) 2018-2023  Robert J. Hijmans
 //
 // This file is part of the "spat" library.
 //
@@ -29,30 +29,35 @@
 #endif
 
 
-SpatRaster::SpatRaster(std::string fname, std::vector<int> subds, std::vector<std::string> subdsname, std::vector<std::string> options) {
+SpatRaster::SpatRaster(std::string fname, std::vector<int> subds, std::vector<std::string> subdsname, std::vector<std::string> drivers, std::vector<std::string> options) {
 #ifdef useGDAL
-	constructFromFile(fname, subds, subdsname, options);
+	constructFromFile(fname, subds, subdsname, drivers, options);
 #endif
 }
 
 
-SpatRaster::SpatRaster(std::vector<std::string> fname, std::vector<int> subds, std::vector<std::string> subdsname, bool multi, std::vector<std::string> options, std::vector<size_t> x) {
+SpatRaster::SpatRaster(std::vector<std::string> fname, std::vector<int> subds, std::vector<std::string> subdsname, bool multi, std::vector<std::string> drivers, std::vector<std::string> options, std::vector<size_t> x) {
 // argument "x" is ignored. It is only there to have four arguments such that the  module
-// can distinguish this constructor from another with three arguments. 
-#ifdef useGDAL
-	if (multi) {
-		constructFromFileMulti(fname[0], subdsname[0], x);
+// can distinguish this constructor from another with three arguments.
+	if (fname.size() == 0) {
+		setError("no filename");
 		return;
 	}
 
-	if (!constructFromFile(fname[0], subds, subdsname, options)) {
-		setError("cannot open file: " + fname[0]);
+#ifdef useGDAL
+	if (multi) {
+		constructFromFileMulti(fname[0], subds, subdsname, drivers, options, x);
+		return;
+	}
+
+	if (!constructFromFile(fname[0], subds, subdsname, drivers, options)) {
+		//setError("cannot open file: " + fname[0]);
 		return;
 	}
 	SpatOptions opt;
 	for (size_t i=1; i<fname.size(); i++) {
 		SpatRaster r;
-		bool ok = r.constructFromFile(fname[i], subds, subdsname, options);
+		bool ok = r.constructFromFile(fname[i], subds, subdsname, drivers, options);
 		if (r.msg.has_warning) {
 			addWarning(r.msg.warnings[0]);
 		}
@@ -74,23 +79,19 @@ SpatRaster::SpatRaster(std::vector<std::string> fname, std::vector<int> subds, s
 }
 
 
-void SpatRaster::setSources(std::vector<SpatRasterSource> s) {
+void SpatRaster::setSources(std::vector<SpatRasterSource> &s) {
 	source = s;
-//	extent = s[0].extent;
-//	srs = s[0].srs;
 }
 
 
-void SpatRaster::setSource(SpatRasterSource s) {
-	s.resize(s.nlyr);
-	std::vector<SpatRasterSource> vs = {s};
-	setSources(vs);
+void SpatRaster::setSource(SpatRasterSource &s) {
+	s.resize(s.nlyr); // appears to be necessary!
+	source = {s};
 }
 
 
-SpatRaster::SpatRaster(SpatRasterSource s) {
-	std::vector<SpatRasterSource> vs = {s};
-	setSources(vs);
+SpatRaster::SpatRaster(SpatRasterSource &s) {
+	source = {s};
 }
 
 
@@ -110,7 +111,7 @@ SpatRaster::SpatRaster() {
 	s.hasValues = false;
 	s.valueType = { 0 };
 	s.layers.resize(1, 0);
-	s.datatype = "";
+	s.dtype = "";
 	s.names = {"lyr.1"};
 	s.srs.proj4 = "+proj=longlat +datum=WGS84";
 	s.srs.wkt = "GEOGCRS[\"WGS 84\", DATUM[\"World Geodetic System 1984\", ELLIPSOID[\"WGS 84\",6378137,298.257223563, LENGTHUNIT[\"metre\",1]]], PRIMEM[\"Greenwich\",0, ANGLEUNIT[\"degree\",0.0174532925199433]], CS[ellipsoidal,2], AXIS[\"geodetic latitude (Lat)\",north, ORDER[1], ANGLEUNIT[\"degree\",0.0174532925199433]], AXIS[\"geodetic longitude (Lon)\",east, ORDER[2], ANGLEUNIT[\"degree\",0.0174532925199433]], USAGE[ SCOPE[\"Horizontal component of 3D system.\"], AREA[\"World.\"], BBOX[-90,-180,90,180]], ID[\"EPSG\",4326]]";
@@ -125,11 +126,11 @@ SpatRaster SpatRaster::dropSource() {
 }
 */
 
-bool SpatRaster::hasValues() { 
+bool SpatRaster::hasValues() {
 //	if (source.size() == 0) {
 //		return false;
 //	} else {
-	return source[0].hasValues ; 
+	return source[0].hasValues ;
 //	}
 }
 
@@ -155,14 +156,17 @@ SpatRaster::SpatRaster(std::vector<unsigned> rcl, std::vector<double> ext, std::
 	//s.layers.resize(1, s.nlyr);
 	//std::iota(s.layers.begin(), s.layers.end(), 0);
 
-	s.datatype = "";
+	s.dtype = "";
 
 #ifdef useGDAL
 	std::string msg;
 	if (!s.srs.set( crs, msg )) {
 		setError(msg);
 		return;
+	} else if (msg != "") {
+		addWarning(msg);
 	}
+
 #else
 	s.srs.proj4 = lrtrim_copy(crs);
 #endif
@@ -193,12 +197,14 @@ SpatRaster::SpatRaster(unsigned nr, unsigned nc, unsigned nl, SpatExtent ext, st
 	s.layers.resize(1, 0);
 	//s.layers.resize(1, _nlyr);
 	//std::iota(s.layers.begin(), s.layers.end(), 0);
-	s.datatype = "";
+	s.dtype = "";
 #ifdef useGDAL
 	std::string msg;
 	if (!s.srs.set(crs, msg )) {
 		setError(msg);
 		return;
+	} else if (msg != "") {
+		addWarning(msg);
 	}
 #else
 	s.srs.proj4 = lrtrim_copy(crs);
@@ -256,6 +262,7 @@ SpatRaster SpatRaster::geometry(long nlyrs, bool properties, bool time, bool uni
 		if (time && hasTime()) {
 			s.hasTime = true;
 			s.timestep = getTimeStep();
+			s.timezone = getTimeZone();
 			s.time = getTime();
 		}
 		if (units && hasUnit()) {
@@ -273,11 +280,23 @@ SpatRaster SpatRaster::geometry(long nlyrs, bool properties, bool time, bool uni
 		out.rgb = rgb;
 		out.rgbtype = rgbtype;
 		out.rgblyrs = rgblyrs;
-	}
+	}	
 	return out;
 }
 
 
+SpatRaster SpatRaster::geometry_opt(long nlyrs, bool properties, bool time, bool units, bool datatype, SpatOptions &opt) {
+
+	if (datatype && hasValues() && (!opt.datatype_set)) {
+		std::vector<std::string> dt = getDataType(true);
+		if ((dt.size() == 1) && (dt[0] != "")) {
+			opt.set_datatype(dt[0]);
+		}
+	}	
+	
+	return geometry(nlyrs, properties, time, units);
+}
+
 SpatRaster SpatRaster::deepCopy() {
 	return *this;
 }
@@ -298,8 +317,9 @@ SpatRaster SpatRaster::setResolution(double xres, double yres) {
 		return(out);
 	}
 	SpatExtent e = getExtent();
-	unsigned nc = round((e.xmax-e.xmin) / xres);
-	unsigned nr = round((e.ymax-e.ymin) / yres);
+	unsigned nc = std::max(1., round((e.xmax-e.xmin) / xres));
+	unsigned nr = std::max(1., round((e.ymax-e.ymin) / yres));
+
 	double xmax = e.xmin + nc * xres;
 	double ymax = e.ymin + nr * yres;
 	unsigned nl = nlyr();
@@ -357,11 +377,16 @@ std::vector<bool> SpatRaster::hasRange() {
 	return(x);
 }
 
-std::vector<int> SpatRaster::getValueType() {
+std::vector<int> SpatRaster::getValueType(bool unique) {
 	std::vector<int> d;
+	d.reserve(nlyr());
 	for (size_t i=0; i<source.size(); i++) {
 		d.insert(d.end(), source[i].valueType.begin(), source[i].valueType.end());
 	}
+	if (unique) {
+		std::sort(d.begin(), d.end());
+		d.erase(std::unique(d.begin(), d.end()), d.end());	
+	}
 	return(d);
 }
 
@@ -410,6 +435,22 @@ bool SpatRaster::is_global_lonlat() {
 	return source[0].srs.is_global_lonlat(e);
 }
 
+int SpatRaster::ns_polar() {
+	int polar = 0;
+	if (!is_lonlat()) {
+		return polar;
+	}
+	SpatExtent e = getExtent();
+	if ((e.ymax - 90) > -0.00001) {
+		polar = 1;
+	}
+	if ((e.ymin + 90) < 0.00001) {
+		polar = polar == 1 ? 2 : -1;
+	}
+	return polar;
+}
+
+
 
 bool SpatRaster::sources_from_file() {
 	for (size_t i=0; i<source.size(); i++) {
@@ -466,9 +507,12 @@ bool SpatRaster::setSRS(std::string crs) {
 	if (!srs.set(crs, msg )) {
 		addWarning("Cannot set raster SRS: "+ msg);
 		return false;
+	} else if (msg != "") {
+		addWarning(msg);
 	}
-	for (size_t i = 0; i < nsrc(); i++) { 
-		source[i].srs = srs; 
+
+	for (size_t i = 0; i < nsrc(); i++) {
+		source[i].srs = srs;
 		if (!source[i].memory) {
 			source[i].parameters_changed = true;
 		}
@@ -484,8 +528,8 @@ bool SpatRaster::setSRS(OGRSpatialReference *poSRS, std::string &msg) {
 		addWarning("Cannot set raster SRS: "+ msg);
 		return false;
 	}
-	for (size_t i = 0; i < nsrc(); i++) { 
-		source[i].srs = srs; 
+	for (size_t i = 0; i < nsrc(); i++) {
+		source[i].srs = srs;
 	}
 	return true;
 }
@@ -586,7 +630,7 @@ bool SpatRaster::setSourceNames(std::vector<std::string> names) {
 
 bool SpatRaster::setNAflag(std::vector<double> flag) {
 	size_t sz = source.size();
-	if (flag.size() == 1) recycle(flag, sz); 
+	if (flag.size() == 1) recycle(flag, sz);
 	if (flag.size() != sz) {
 		return false;
 	}
@@ -623,9 +667,9 @@ std::vector<double> SpatRaster::getNAflag() {
 
 
 bool SpatRaster::hasTime() {
-	bool test = source[0].hasTime; 
+	bool test = source[0].hasTime;
 	for (size_t i=1; i<source.size(); i++) {
-		test = test && source[i].hasTime; 
+		test = test && source[i].hasTime;
 	}
 	return test;
 }
@@ -650,13 +694,13 @@ std::vector<std::string> SpatRaster::getTimeStr(bool addstep) {
 		for (size_t i=0; i < time.size(); i++) {
 			std::vector<int> x = get_date(time[i]);
 			if (x.size() > 2) {
-				out.push_back( std::to_string(x[0]) + "-" 
+				out.push_back( std::to_string(x[0]) + "-"
 						  + std::to_string(x[1]) + "-"
 						  + std::to_string(x[2]) + " "
 						  + std::to_string(x[3]) + ":"
 						  + std::to_string(x[4]) + ":"
 						  + std::to_string(x[5]) );
-						  
+
 			} else {
 				out.push_back("");
 			}
@@ -665,10 +709,10 @@ std::vector<std::string> SpatRaster::getTimeStr(bool addstep) {
 		for (size_t i=0; i < time.size(); i++) {
 			std::vector<int> x = get_date(time[i]);
 			if (x.size() > 2) {
-				out.push_back( std::to_string(x[0]) + "-" 
+				out.push_back( std::to_string(x[0]) + "-"
 						  + std::to_string(x[1]) + "-"
 						  + std::to_string(x[2]) );
-						  
+
 			} else {
 				out.push_back("");
 			}
@@ -699,12 +743,17 @@ std::string SpatRaster::getTimeStep() {
 	return source[0].timestep;
 }
 
-bool SpatRaster::setTime(std::vector<int_64> time, std::string step) {
+std::string SpatRaster::getTimeZone() {
+	return source[0].timezone;
+}
 
-	if (time.size() == 0 || step == "remove") { 
+bool SpatRaster::setTime(std::vector<int_64> time, std::string step, std::string zone) {
+
+	if (time.size() == 0 || step == "remove") {
 		for (size_t i=0; i<source.size(); i++)	{
 			source[i].time = std::vector<int_64> (source[i].nlyr);
 			source[i].timestep = "";
+			source[i].timezone = "";
 			source[i].hasTime = false;
 		}
 		return true;
@@ -712,15 +761,18 @@ bool SpatRaster::setTime(std::vector<int_64> time, std::string step) {
 
 	if (time.size() != nlyr()) {
 		return false;
-	} 
-	if (!((step == "seconds") || (step == "raw")  || (step == "days"))) {  // "months", "years"
+	}
+	
+	std::vector<std::string> steps = {"seconds", "raw", "days", "yearmonths", "years", "months"};
+	if (!is_in_vector(step, steps)) {
 		return false;
-	} 
+	}
 	size_t begin=0;
 	for (size_t i=0; i<source.size(); i++)	{
 		size_t end = begin + source[i].nlyr;
         source[i].time = std::vector<int_64> (time.begin() + begin, time.begin() + end);
 		source[i].timestep = step;
+		source[i].timezone = zone;
 		source[i].hasTime = true;
         begin = end;
     }
@@ -775,9 +827,13 @@ bool SpatRaster::setDepth(std::vector<double> depths) {
 
 bool SpatRaster::setUnit(std::vector<std::string> units) {
 	if (units.size() == 1) {
+		bool hu = true;
+		if (units[0] == "") {
+			hu = false;
+		}
         for (size_t i=0; i<source.size(); i++)	{
             source[i].unit = std::vector<std::string> (source[i].nlyr, units[0]);
-			source[i].hasUnit = true;
+			source[i].hasUnit = hu;
         }
         return true;
 	} else if (units.size() != nlyr()) {
@@ -795,9 +851,9 @@ bool SpatRaster::setUnit(std::vector<std::string> units) {
 }
 
 bool SpatRaster::hasUnit() {
-	bool test = source[0].hasUnit; 
+	bool test = source[0].hasUnit;
 	for (size_t i=1; i<source.size(); i++) {
-		test = test && source[i].hasUnit; 
+		test = test && source[i].hasUnit;
 	}
 	return test;
 }
@@ -824,7 +880,7 @@ double SpatRaster::xres() {
 	return (extent.xmax - extent.xmin) / ncol() ;
 }
 
-double SpatRaster::yres() { 
+double SpatRaster::yres() {
 	SpatExtent extent = getExtent();
 	return (extent.ymax - extent.ymin) / nrow() ;
 }
@@ -832,8 +888,8 @@ double SpatRaster::yres() {
 
 bool SpatRaster::valid_sources(bool files, bool rotated) {
 	std::vector<std::string> ff;
-	for (size_t i=0; i<source.size(); i++) { 
-		std::string f = source[i].filename; 
+	for (size_t i=0; i<source.size(); i++) {
+		std::string f = source[i].filename;
 		if (f == "") continue;
 		if (files) {
 			std::size_t found = f.find(":"); // perhaps http: or PG:xxx
@@ -870,7 +926,7 @@ bool SpatRaster::removeWindow() {
 	for (size_t i=0; i<nsrc(); i++) {
 		if (source[i].hasWindow) {
 			SpatExtent e = source[0].window.full_extent;
-			setExtent(e, true, "");
+			setExtent(e, true, true, "");
 			for (size_t i=0; i<source.size(); i++) {
 				source[i].hasWindow = false;
 				source[i].nrow = source[0].window.full_nrow;
@@ -887,7 +943,7 @@ bool SpatRaster::setWindow(SpatExtent x) {
 	if ( !x.valid() ) {
 		setError("invalid extent");
 		return false;
-	} 
+	}
 
 	removeWindow();
 	x = align(x, "near");
@@ -900,7 +956,7 @@ bool SpatRaster::setWindow(SpatExtent x) {
 	if ( !e.valid() ) {
 		setError("extents do not overlap");
 		return false;
-	} 
+	}
 
 // get read-window
 	double xr = xres();
@@ -936,7 +992,7 @@ bool SpatRaster::setWindow(SpatExtent x) {
 	if (r < 0) {
 		expand = true;
 		exp[3] = trunc(abs(x.xmin - e.xmin) / xres());
-	} 
+	}
 
 	if (expand) {
 		setError("expansion is not yet allowed");
@@ -949,11 +1005,11 @@ bool SpatRaster::setWindow(SpatExtent x) {
 		source[i].window.expand = exp;
 		source[i].window.expanded  = expand;
 		source[i].window.full_extent = getExtent();
-		source[i].window.full_nrow   = source[i].nrow;
-		source[i].window.full_ncol   = source[i].ncol;
-		source[i].hasWindow     = true;
+		source[i].window.full_nrow = source[i].nrow;
+		source[i].window.full_ncol = source[i].ncol;
+		source[i].hasWindow = true;
 	}
-	setExtent(x, true, "");
+	setExtent(x, true, true, "");
 
 	return true;
 }
@@ -1000,7 +1056,7 @@ SpatRaster SpatRaster::makeCategorical(unsigned layer, SpatOptions &opt) {
 
 	if (!hasValues()) {
 		SpatRaster out;
-		out.setError("cannot make categries if the raster has no values");
+		out.setError("cannot make categories if the raster has no values");
 		return out;
 	}
 
@@ -1010,7 +1066,7 @@ SpatRaster SpatRaster::makeCategorical(unsigned layer, SpatOptions &opt) {
 
 	r.math2("round", 0, fopt);
 
-	std::vector<std::vector<double>> u = r.unique(false, fopt);
+	std::vector<std::vector<double>> u = r.unique(false, true, fopt);
 /*
 	std::vector<double> id(u[0].size());
 	std::iota(id.begin(), id.end(), 0);
@@ -1044,13 +1100,13 @@ SpatRaster SpatRaster::makeCategorical(unsigned layer, SpatOptions &opt) {
 
 
 bool SpatRaster::createCategories(unsigned layer, SpatOptions &opt) {
-	if (layer > (nlyr()-1)) { 
+	if (layer > (nlyr()-1)) {
 		setError("invalid layer number");
 		return(false);
 	}
 	std::vector<unsigned> lyrs(1, layer);
 	SpatRaster r = subset(lyrs, opt);
-	std::vector<std::vector<double>> u = r.unique(false, opt);
+	std::vector<std::vector<double>> u = r.unique(false, true, opt);
     std::vector<unsigned> sl = findLyr(layer);
 
 	std::vector<std::string> s(u[0].size());
@@ -1070,32 +1126,55 @@ bool SpatRaster::createCategories(unsigned layer, SpatOptions &opt) {
 
 
 std::vector<bool> SpatRaster::hasCategories() {
-	std::vector<bool> b(nlyr());
+	std::vector<bool> b;
+	b.reserve(nlyr());
 	std::vector<unsigned> ns = nlyrBySource();
-	unsigned k = 0;
 	for (size_t i=0; i<ns.size(); i++) {
 		for (size_t j=0; j<ns[i]; j++) {
-			b[k] = source[i].hasCategories[j];
-			k++;
+			b.push_back(source[i].hasCategories[j]);
 		}
 	}
 	return b;
 }
 
+std::vector<std::string> SpatRaster::getDataType(bool unique) {
+	std::vector<std::string> d;
+	size_t n = nsrc();
+	d.reserve(n);
+	for (size_t i=0; i<n; i++) {
+		d.push_back(source[i].dtype);
+	}
+	if (unique) {
+		std::sort(d.begin(), d.end());
+		d.erase(std::unique(d.begin(), d.end()), d.end());
+	}
+	return d;
+}
+
+std::vector<std::string> SpatRaster::dataType() {
+	std::vector<std::string> d;
+	size_t n = nsrc();
+	d.reserve(n);
+	for (size_t i=0; i<n; i++) {
+		d.push_back(source[i].dtype);
+	}
+	return d;
+}
+
 
 
 
 bool SpatRaster::setLabels(unsigned layer, std::vector<long> values, std::vector<std::string> labels, std::string name) {
 
-	if (layer > (nlyr()-1)) { 
+	if (layer > (nlyr()-1)) {
 		setError("invalid layer number");
 		return(false);
 	}
-	if (values.size() != labels.size()) { 
+	if (values.size() != labels.size()) {
 		setError("size of values is not the same as the size of labels");
 		return(false);
 	}
-	if (values.size() == 0) { 
+	if (values.size() == 0) {
 		addWarning("no labels");
 		return(true);
 	}
@@ -1118,9 +1197,10 @@ bool SpatRaster::setLabels(unsigned layer, std::vector<long> values, std::vector
 
 
 
+
 bool SpatRaster::setCategories(unsigned layer, SpatDataFrame d, unsigned index) {
 
-	if (layer > (nlyr()-1)) { 
+	if (layer >= nlyr()) {
 		setError("invalid layer number");
 		return(false);
 	}
@@ -1139,15 +1219,24 @@ bool SpatRaster::setCategories(unsigned layer, SpatDataFrame d, unsigned index)
 }
 
 
-bool SpatRaster::removeCategories(unsigned layer) {
-	if (layer > (nlyr()-1)) { 
+bool SpatRaster::removeCategories(long layer) {
+	if (layer > (((long)nlyr())-1)) {
 		setError("invalid layer number");
 		return(false);
 	}
-    std::vector<unsigned> sl = findLyr(layer);
 	SpatCategories s;
-	source[sl[0]].cats[sl[1]] = s;
-	source[sl[0]].hasCategories[sl[1]] = false;
+	if (layer < 0) {
+		for (size_t i=0; i<source.size(); i++) {
+			for (size_t j=0; j<source[i].cats.size(); j++) {
+				source[i].cats[j] = s;
+				source[i].hasCategories[j] = false;
+			}
+		}
+	} else {
+		std::vector<unsigned> sl = findLyr(layer);
+		source[sl[0]].cats[sl[1]] = s;
+		source[sl[0]].hasCategories[sl[1]] = false;
+	}
 	return true;
 }
 
@@ -1159,6 +1248,7 @@ SpatCategories SpatRaster::getLayerCategories(unsigned layer) {
 
 std::vector<SpatCategories> SpatRaster::getCategories() {
 	std::vector<SpatCategories> cats;
+	cats.reserve(nlyr());
 	for (size_t i=0; i<source.size(); i++) {
 		cats.insert(cats.end(), source[i].cats.begin(), source[i].cats.end());
 	}
@@ -1166,6 +1256,73 @@ std::vector<SpatCategories> SpatRaster::getCategories() {
 }
 
 
+std::vector<std::vector<double>> SpatRaster::getScaleOffset() {
+	std::vector<std::vector<double>> so(2);
+	so[0].reserve(nlyr());
+	so[1].reserve(nlyr());
+	for (size_t i=0; i<source.size(); i++) {
+		so[0].insert(so[0].end(), source[i].scale.begin(), source[i].scale.end());
+		so[1].insert(so[1].end(), source[i].offset.begin(), source[i].offset.end());
+	}
+	return so;
+}
+
+bool SpatRaster::setScaleOffset(std::vector<double> sc, std::vector<double> of) {
+	size_t n = sc.size();
+	size_t nl = nlyr();
+	if (n != of.size()) {
+		setError("length of scale and offset should be the same");
+		return false;
+	}
+	if (n > nl) {
+		setError("length of scale and offset cannot exceed the number of layers");
+		return false;
+	}
+	if (n < nl) {
+		recycle(sc, nl);
+		recycle(of, nl);
+		if (n > 1) {
+			addWarning("recycling scale and offset to the number of layers");
+		}
+	}
+	size_t k=0;
+	size_t nc=ncell();
+	for (size_t i=0; i<source.size(); i++)	{
+		if (source[i].memory) {
+			for (size_t j=0; j<source[i].nlyr; j++) {
+				size_t loff = j * nc;
+				if ((sc[k] != 1) || (of[k] != 0)) {
+					for (size_t p=loff; p<(loff+nc); p++) {
+						source[i].values[p] = source[i].values[p] * sc[k] + of[k];
+					}
+					source[i].range_min[j] = source[i].range_min[j] * sc[k] + of[k];
+					source[i].range_max[j] = source[i].range_max[j] * sc[k] + of[k];
+				}
+				k++;
+			}
+		} else {
+			for (size_t j=0; j<source[i].nlyr; j++) {
+				if (source[i].has_scale_offset[j]) {
+					source[i].range_min[j] = (source[i].range_min[j] - source[i].offset[j]) / source[i].scale[j];
+					source[i].range_max[j] = (source[i].range_max[j] - source[i].offset[j]) / source[i].scale[j];
+				}
+				source[i].scale[j] = sc[k];
+				source[i].offset[j] = of[k];
+				if ((sc[k] != 1) || (of[k] != 0)) {
+					source[i].has_scale_offset[j] = true;
+					source[i].range_min[j] = source[i].range_min[j] * sc[k] + of[k];
+					source[i].range_max[j] = source[i].range_max[j] * sc[k] + of[k];
+				} else {
+					source[i].has_scale_offset[j] = false;
+				}
+				k++;
+			}
+		}
+	}
+	return true;
+}
+
+
 std::vector<std::string> SpatRaster::getLabels(unsigned layer) {
 	std::vector<std::string> out;
 	if (layer >= nlyr()) return out;
@@ -1185,7 +1342,7 @@ std::vector<std::string> SpatRaster::getLabels(unsigned layer) {
 }
 
 bool SpatRaster::setCatIndex(unsigned layer, unsigned index) {
-	if (layer > (nlyr()-1)) { 
+	if (layer > (nlyr()-1)) {
 		return(false);
 	}
     std::vector<unsigned> sl = findLyr(layer);
@@ -1200,13 +1357,58 @@ bool SpatRaster::setCatIndex(unsigned layer, unsigned index) {
 }
 
 int SpatRaster::getCatIndex(unsigned layer) {
-	if (layer > (nlyr()-1)) { 
+	if (layer > (nlyr()-1)) {
 		return(-1);
 	}
     std::vector<unsigned> sl = findLyr(layer);
 	return source[sl[0]].cats[sl[1]].index;
 }
 
+SpatRaster SpatRaster::dropLevels() {
+	std::vector<bool> hascats = hasCategories();
+	bool bany = false;
+	for (size_t i=0; i<hascats.size(); i++) {
+		if (hascats[i]) {
+			bany = true;
+			break;
+		}
+	}
+	if (!bany) return *this;
+
+	std::vector<SpatCategories> cats = getCategories();
+	SpatOptions opt;
+	SpatRaster out = *this;
+	std::vector<std::vector<double>> uvv = unique(true, true, opt);
+	for (size_t i=0; i<hascats.size(); i++) {
+		if (hascats[i]) {
+			SpatCategories lyrcats = cats[i];
+			size_t n = lyrcats.d.nrow();
+			std::vector<double> uv = uvv[i];
+			std::vector<long> uvi(uv.size());
+			for (size_t j=0; j<uv.size(); j++) {
+				uvi[j] = uv[j];
+			}
+			std::vector<long> isin;
+			isin.reserve(n);
+			for (size_t j=0; j<n; j++) {
+				for (size_t k=0; k<uvi.size(); k++) {
+					if (lyrcats.d.iv[0][j] == uvi[k]) {
+						isin.push_back(j);
+						continue;
+					}
+				}
+			}
+			lyrcats.d = lyrcats.d.subset_rows(isin);
+			if (lyrcats.d.nrow() == 0) {
+				out.removeCategories(i);
+			} else {
+				out.setCategories(i, lyrcats.d, lyrcats.index);
+			}
+		}
+	}
+	return out;
+}
+
 
 
 std::vector<SpatDataFrame> SpatRaster::getColors() {
@@ -1223,9 +1425,6 @@ bool SpatRaster::setColors(size_t layer, SpatDataFrame cols) {
 	if (cols.ncol() < 4 || cols.ncol() > 5) {
 		return false;
 	}
-//	if (cols.nrow() != 256) {
-//		return false;
-//	}
 	if (layer >= nlyr()) {
 		return false;
 	}
@@ -1243,7 +1442,7 @@ bool SpatRaster::setColors(size_t layer, SpatDataFrame cols) {
 	}
 
 	source[sl[0]].cols[sl[1]] = cols;
-	source[sl[0]].hasColors[sl[1]] = (cols.nrow() > 1);
+	source[sl[0]].hasColors[sl[1]] = (cols.nrow() > 0);
 	return true;
 }
 
@@ -1352,7 +1551,7 @@ std::vector<double> SpatRaster::cellFromRowColCombine(std::vector<int_64> row, s
 	for (size_t i=0; i<n; i++) {
 		for (size_t j=0; j<m; j++) {
 			if (row[i] < 0 || row[i] >= nr || col[j]<0 || col[j] >= nc) {
-				x.push_back(NAN);				
+				x.push_back(NAN);
 			} else {
 				x.push_back(row[i] * nc + col[j]);
 			}
@@ -1505,8 +1704,8 @@ std::vector<std::vector<int_64>> SpatRaster::rowColFromCell(std::vector<double>
 
 std::vector<std::vector<int_64>>  SpatRaster::rowColFromExtent(SpatExtent e) {
 	std::vector<std::vector<double>> xy = e.asPoints();
-	std::vector<int_64> col = colFromX(xy[0]); 
-	std::vector<int_64> row = rowFromY(xy[1]); 
+	std::vector<int_64> col = colFromX(xy[0]);
+	std::vector<int_64> row = rowFromY(xy[1]);
 	std::vector<std::vector<int_64>> out = { row, col };
 	return out;
 }
@@ -1655,7 +1854,7 @@ std::vector<double> SpatRaster::adjacent(std::vector<double> cells, std::string
             rows = {r[i]-2, r[i]-2, r[i]-1, r[i]-1, r[i]-1, r[i]-1, r[i]-1, r[i]  , r[i]  , r[i]+1, r[i]+1, r[i]+1, r[i]+1, r[i]+1, r[i]+2, r[i]+2};
             cols = {c[i]-1, c[i]+1, c[i]-2, c[i]-1, c[i],   c[i]+1, c[i]+2, c[i]-1, c[i]+1, c[i]-2, c[i]-1, c[i]  , c[i]+1, c[i]+2, c[i]-1, c[i]+1};
             if (globlatlon) {
-                if ((c[i]==0) | (c[i]==1)) {
+                if ((c[i]==0) || (c[i]==1)) {
                     for (size_t j=0; j<16; j++) {
                         cols[j] = (cols[j] < 0) ? nc-cols[j] : cols[j];
                     }
@@ -1676,24 +1875,98 @@ std::vector<double> SpatRaster::adjacent(std::vector<double> cells, std::string
 }
 
 
-SpatVector SpatRaster::as_points(bool values, bool narm, SpatOptions &opt) {
+SpatVector SpatRaster::as_multipoints(bool narm, bool nall, SpatOptions &opt) {
 
 	BlockSize bs = getBlockSize(opt);
-	std::vector<double> v, vout;
-	vout.reserve(v.size());
+    size_t ncl = ncell();
 	SpatVector pv;
-	SpatGeom g;
-	g.gtype = points;
+	pv.reserve(1);
+
+    std::vector<std::vector<double>> xy;
+	if (!narm) {
+        for (size_t i=0; i<ncl; i++) {
+            xy = xyFromCell(i);
+			SpatPart p(xy[0], xy[1]);
+			SpatGeom g(p, points);
+			pv.addGeom(g);
+			g.parts.resize(0);
+        }
+		return pv;
+	}
+
+	if (!readStart()) {
+		pv.setError(getError());
+		return(pv);
+	}
+
+	size_t nc = ncol();
+	unsigned nl = nlyr();
+	std::vector<double> v, x, y;
+	for (size_t i = 0; i < bs.n; i++) {
+		readValues(v, bs.row[i], bs.nrows[i], 0, nc);
+        size_t off1 = (bs.row[i] * nc);
+ 		size_t vnc = bs.nrows[i] * nc;
+		for (size_t j=0; j<vnc; j++) {
+			if (nall) {
+				bool allna = true;
+				for (size_t lyr=0; lyr<nl; lyr++) {
+					size_t off2 = lyr*vnc;
+					if (!std::isnan(v[off2+j])) {
+						allna = false;
+						continue;
+					}
+				}
+				if (allna) continue;
+			} else {
+				bool foundna = false;
+				for (size_t lyr=0; lyr<nl; lyr++) {
+					size_t off2 = lyr*vnc;
+					if (std::isnan(v[off2+j])) {
+						foundna = true;
+						continue;
+					}
+				}
+				if (foundna) continue;
+			}
+			xy = xyFromCell( off1+j );
+			x.push_back(xy[0][0]);
+			y.push_back(xy[1][0]);
+		}
+	}
+	SpatPart p(x, y);
+	SpatGeom g(p, points);
+	pv.addGeom(g);
+
+	readStop();
+	pv.srs = source[0].srs;
+	return(pv);
+}
+
+
+SpatVector SpatRaster::as_points(bool values, bool narm, bool nall, SpatOptions &opt) {
+
+	BlockSize bs = getBlockSize(opt);
+    size_t ncl = ncell();
+	SpatVector pv;
+	pv.reserve(ncl);
+	pv.srs = source[0].srs;
+
+	if (!hasValues()) {
+		if (values) {
+			pv.addWarning("raster has no values");
+		}
+		values = false;
+		narm = false;
+	}
 
     std::vector<std::vector<double>> xy;
 	if ((!values) && (!narm)) {
-        double nc = ncell();
-        for (size_t i=0; i<nc; i++) {
+        for (size_t i=0; i<ncl; i++) {
             xy = xyFromCell(i);
 			SpatPart p(xy[0], xy[1]);
-			g.addPart(p);
+			SpatGeom g(p, points);
 			pv.addGeom(g);
-			g.parts.resize(0);
+			//g.parts.resize(0);
         }
 		return pv;
 	}
@@ -1711,26 +1984,43 @@ SpatVector SpatRaster::as_points(bool values, bool narm, SpatOptions &opt) {
 
 	size_t nc = ncol();
 	unsigned nl = nlyr();
+	std::vector<double> v;
 	for (size_t i = 0; i < bs.n; i++) {
 		readValues(v, bs.row[i], bs.nrows[i], 0, nc);
         size_t off1 = (bs.row[i] * nc);
  		size_t vnc = bs.nrows[i] * nc;
 		if (narm) {
+			if (values) {
+				pv.df.reserve(ncl);
+			}
 			for (size_t j=0; j<vnc; j++) {
-				bool foundna = false;
-				for (size_t lyr=0; lyr<nl; lyr++) {
-                    size_t off2 = lyr*vnc;
-                    if (std::isnan(v[off2+j])) {
-                        foundna = true;
-                        continue;
-                    }
-                }
-                if (foundna) continue;
+				if (nall) {
+					bool allna = true;
+					for (size_t lyr=0; lyr<nl; lyr++) {
+						size_t off2 = lyr*vnc;
+						if (!std::isnan(v[off2+j])) {
+							allna = false;
+							continue;
+						}
+					}
+					if (allna) continue;
+				} else {
+					bool foundna = false;
+					for (size_t lyr=0; lyr<nl; lyr++) {
+						size_t off2 = lyr*vnc;
+						if (std::isnan(v[off2+j])) {
+							foundna = true;
+							continue;
+						}
+					}
+					if (foundna) continue;
+				}
+
+
                 xy = xyFromCell( off1+j );
                 SpatPart p(xy[0], xy[1]);
-                g.addPart(p);
+                SpatGeom g(p, points);
                 pv.addGeom(g);
-                g.parts.resize(0);
                 if (values) {
                     for (size_t lyr=0; lyr<nl; lyr++) {
                         unsigned off2 = lyr*vnc;
@@ -1742,9 +2032,8 @@ SpatVector SpatRaster::as_points(bool values, bool narm, SpatOptions &opt) {
 			for (size_t j=0; j<vnc; j++) {
                 xy = xyFromCell(off1+j);
                 SpatPart p(xy[0], xy[1]);
-                g.addPart(p);
+                SpatGeom g(p, points);
                 pv.addGeom(g);
-                g.parts.resize(0);
 			}
             for (size_t lyr=0; lyr<nl; lyr++) {
 				size_t off2 = lyr*vnc;
@@ -1753,12 +2042,128 @@ SpatVector SpatRaster::as_points(bool values, bool narm, SpatOptions &opt) {
 		}
 	}
 	readStop();
-	pv.srs = source[0].srs;
+//	pv.srs = source[0].srs;
 	return(pv);
 }
 
+std::vector<std::vector<double>> SpatRaster::as_points_value(const double& target, SpatOptions &opt) {
+
+	std::vector<std::vector<double>> xy(2);
+	if (nlyr() > 1) {
+		setError("can only process one layer");
+		return xy;
+	}
+
+	BlockSize bs = getBlockSize(opt);
+
+	if (!readStart()) {
+		return(xy);
+	}
+
+	size_t nc = ncol();
+    size_t ncl = ncell();
+	std::vector<double> cells;
+	cells.reserve(std::min(ncl/10, (size_t)10000));
+
+	std::vector<double> v;
+	if (std::isnan(target)) {
+		for (size_t i = 0; i < bs.n; i++) {
+			readValues(v, bs.row[i], bs.nrows[i], 0, nc);
+			size_t base = (bs.row[i] * nc);
+			size_t szv = v.size();
+			for (size_t j=0; j<szv; j++) {
+				if (std::isnan(v[j])) {
+					cells.push_back(base+j);
+				}
+			}
+		}
+	} else {
+		for (size_t i = 0; i < bs.n; i++) {
+			readValues(v, bs.row[i], bs.nrows[i], 0, nc);
+			size_t base = (bs.row[i] * nc);
+			size_t szv = v.size();
+			for (size_t j=0; j<szv; j++) {
+				if (v[j] == target) {
+					cells.push_back(base+j);
+				}
+			}
+		}
+	}
+	readStop();
+	return xyFromCell(cells);
+}
+
+std::vector<std::vector<double>> SpatRaster::cells_notna(SpatOptions &opt) {
+
+	std::vector<std::vector<double>> out(2);
+	if (nlyr() > 1) {
+		setError("can only process one layer");
+		return out;
+	}
+
+	BlockSize bs = getBlockSize(opt);
+
+	if (!readStart()) {
+		return(out);
+	}
+
+	size_t nc = ncol();
+    size_t ncl = ncell();
+	size_t rs = std::max(ncl/50, (size_t)10000);
+	out[0].reserve(rs);
+	out[1].reserve(rs);
+
+	for (size_t i = 0; i < bs.n; i++) {
+		std::vector<double> v;
+		readValues(v, bs.row[i], bs.nrows[i], 0, nc);
+		size_t base = (bs.row[i] * nc);
+		size_t szv = v.size();
+		for (size_t j=0; j<szv; j++) {
+			if (!std::isnan(v[j])) {
+				out[0].push_back(base+j); // cell
+				out[1].push_back(v[j]); // value
+			}
+		}
+	}
+	readStop();
+	return out;
+}
+
+std::vector<double> SpatRaster::cells_notna_novalues(SpatOptions &opt) {
 
 
+	if (nlyr() > 1) {
+		SpatOptions topt(opt);
+		SpatRaster x = nonan(true, topt);
+		return x.cells_notna_novalues(opt);
+	}
+	
+	std::vector<double> out;
+	BlockSize bs = getBlockSize(opt);
+
+	if (!readStart()) {
+		return(out);
+	}
+
+	size_t nc = ncol();
+    size_t ncl = ncell();
+	size_t rs = std::max(ncl/500, (size_t)10000);
+	out.reserve(rs);
+
+	for (size_t i = 0; i < bs.n; i++) {
+		std::vector<double> v;
+		readValues(v, bs.row[i], bs.nrows[i], 0, nc);
+		size_t base = (bs.row[i] * nc);
+		size_t szv = v.size();
+		for (size_t j=0; j<szv; j++) {
+			if (!std::isnan(v[j])) {
+				out.push_back(base+j); // cell
+			}
+		}
+	}
+	readStop();
+	return out;
+}
 
 
 void getCorners(std::vector<double> &x,  std::vector<double> &y, const double &X, const double &Y, const double &xr, const double &yr) {
@@ -1832,7 +2237,7 @@ SpatVector SpatRaster::as_polygons(bool values, bool narm) {
 
 */
 
-SpatVector SpatRaster::as_polygons(bool trunc, bool dissolve, bool values, bool narm, SpatOptions &opt) {
+SpatVector SpatRaster::as_polygons(bool trunc, bool dissolve, bool values, bool narm, bool nall, SpatOptions &opt) {
 
 	if (!hasValues()) {
 		values = false;
@@ -1882,13 +2287,25 @@ SpatVector SpatRaster::as_polygons(bool trunc, bool dissolve, bool values, bool
 	std::vector<double> cells(ncell()) ;
 	std::iota (std::begin(cells), std::end(cells), 0);
 	std::vector< std::vector<double> > xy = xyFromCell(cells);
+	vect.reserve(cells.size());
 	for (int i=nc-1; i>=0; i--) {
 		if (narm) {
-			bool erase = false;
-			for (size_t j=0; j<nl; j++) {
-				if (std::isnan(vect.df.dv[j][i])) {
-					erase=true;
-					break;
+			bool erase;
+			if (nall) {
+				erase = true;
+				for (size_t j=0; j<nl; j++) {
+					if (!std::isnan(vect.df.dv[j][i])) {
+						erase=false;
+						break;
+					}
+				}
+			} else {
+				erase = false;
+				for (size_t j=0; j<nl; j++) {
+					if (std::isnan(vect.df.dv[j][i])) {
+						erase=true;
+						break;
+					}
 				}
 			}
 			if (erase) {
@@ -2008,7 +2425,7 @@ void SpatRaster::removeRGB(){
 
 
 bool SpatRaster::to_memory(SpatOptions &opt) {
-	if ((nsrc() == 1) & (source[0].memory)) {
+	if ((nsrc() == 1) && (source[0].memory)) {
 		return true;
 	}
 	SpatRaster g = geometry();
@@ -2032,6 +2449,7 @@ SpatRaster SpatRaster::to_memory_copy(SpatOptions &opt) {
 
 std::vector<int> SpatRaster::getFileBlocksize() {
 	std::vector<int> b;
+	b.reserve(2 * nlyr());
 	for (size_t i=0; i<source.size(); i++) {
 		b.insert(b.end(), source[i].blockrows.begin(), source[i].blockrows.end());
 	}
